#!/usr/bin/env python
import sys, os
sys.dont_write_bytecode = True

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.cluster import KMeans

import modules.conditions as conditions
import modules.setting as setting
import modules.converter as converter
import modules.analysis as analysis

import random
random.seed(1107)
np.random.seed(1107)

import warnings
warnings.filterwarnings('ignore')

import argparse
parser = argparse.ArgumentParser()
parser.add_argument('-ex', '--extract', type=str, default='Rh')
parser.add_argument('-wt', '--extract_wt', type=float, default=-1)
parser.add_argument('-r', '--extract_range', type=str, default='equal') # 'small', 'equal', or 'large'

args = vars(parser.parse_args())

######################################################
######################################################
plt.rcParams['font.size'] = 18
plt.rcParams['xtick.labelsize'] = 17
plt.rcParams['ytick.labelsize'] = 17
plt.rcParams['font.family'] = 'DejaVu Sans'
plt.rcParams['font.sans-serif'] = ['Arial']
plt.rcParams["mathtext.fontset"] = "stixsans"
plt.rcParams.update({'mathtext.default':'regular'})
plt.rcParams['xtick.direction'] = 'out'
plt.rcParams['ytick.direction'] = 'out'
plt.rcParams['xtick.major.width'] = 1.0
plt.rcParams['ytick.major.width'] = 1.0
plt.rcParams['axes.linewidth'] = 1.0
plt.rcParams['axes.grid'] = False
plt.rcParams['grid.linestyle'] = '--'
plt.rcParams['grid.linewidth'] = 2
plt.rcParams["legend.markerscale"] = 2
plt.rcParams["legend.fancybox"] = False
plt.rcParams["legend.framealpha"] = 1
plt.rcParams["legend.edgecolor"] = 'black'
######################################################
######################################################

PATH = 'extraction'
os.makedirs(PATH, exist_ok = True)


extract = args['extract']
extract_wt = args['extract_wt']
extract_range = args['extract_range']

### Basic Settings ###
condition = conditions.calc_condition()
date, Reaction = condition['date'], condition['Reaction']
ML_model, Search_method = condition['ML_model'], condition['Search_method']
catal_number, K_cluster = condition['catal_number'], condition['K_cluster']
target_temp = condition['target_temp']

pgm_model = condition['pgm_model']
add_model = condition['add_model']
supp_model = condition['supp_model']

pgm_num = condition['pgm_num']
add_num = condition['add_num']
supp_num = condition['supp_num']

cand_pgm_num = condition['cand_pgm_num']
cand_add_num = condition['cand_add_num']
cand_supp_num = condition['cand_supp_num']

data_file_name = f'data/{date}_{Reaction}_dataset.xlsx'
data_sheet_name = condition['data_sheet_name']
skip_rows = condition['skip_rows']
desc_file_name = f'data/Descriptors_{Reaction}.xlsx'
elem_desc_sheet = 'Descriptors_elem'
supp_desc_sheet = 'Descriptors_supp'


### Define feat and target ###
data_cols = setting.data_columns(condition)
desc_cols = setting.desc_columns(Reaction)

converted = converter.data_convert(condition, data_cols, desc_cols,
                                   sheet_name='data_sheet_name',
                                   select_desc=True, use_models=[],
                                   local=False, idx=None)

elem_lists = converter.elem_list(condition, converted, desc_cols)
pgm_list = elem_lists['use_pgm']
add_list = elem_lists['use_add']
supp_list = elem_lists['use_supp']

### INPUT FILES (About Cand.), Top catalysts OUTPUT Files, and K-mean catalysts OUTPUT Files ###
if pgm_num != 0 and add_num != 0 and supp_num != 0:
    cand_data = f'results/{date}_{Reaction}_cand_sum_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}.csv'
    out_csv_file_top = f'{PATH}/{date}_{Reaction}_cand_sum_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_{extract}.csv'
    fig_title_top = f'{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_top{catal_number}_{extract}'
    out_fig_name_top = f'{PATH}/{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_top{catal_number}_{extract}.png'
    out_csv_file_K = f"{PATH}/{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_K{K_cluster}_{extract}.csv"
    fig_title_K = f'{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_K{K_cluster}_{extract}'
    out_fig_name_K = f"{PATH}/{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_K{K_cluster}_{extract}.png"

elif pgm_num == 0 and add_num != 0 and supp_num != 0:
    cand_data = f'results/{date}_{Reaction}_cand_sum_{ML_model}_prop{add_model}{supp_model}_{Search_method}.csv'
    out_csv_file_top = f'{PATH}/{date}_{Reaction}_cand_sum_{ML_model}_prop{add_model}{supp_model}_{Search_method}_{extract}.csv'
    fig_title_top = f'{date}_{Reaction}_{ML_model}_prop{add_model}{supp_model}_{Search_method}_top{catal_number}_{extract}'
    out_fig_name_top = f'{PATH}/{date}_{Reaction}_{ML_model}_prop{add_model}{supp_model}_{Search_method}_top{catal_number}_{extract}.png'
    out_csv_file_K = f"{PATH}/{date}_{Reaction}_{ML_model}_prop{add_model}{supp_model}_{Search_method}_K{K_cluster}_{extract}.csv"
    fig_title_K = f'{date}_{Reaction}_{ML_model}_prop{add_model}{supp_model}_{Search_method}_K{K_cluster}_{extract}'
    out_fig_name_K = f"{PATH}/{date}_{Reaction}_{ML_model}_prop{add_model}{supp_model}_{Search_method}_K{K_cluster}_{extract}.png"

elif pgm_num == 0 and add_num != 0 and supp_num == 0:
    cand_data = f'results/{date}_{Reaction}_cand_sum_{ML_model}_prop{add_model}_{Search_method}.csv'
    out_csv_file_top = f'{PATH}/{date}_{Reaction}_cand_sum_{ML_model}_prop{add_model}_{Search_method}_{extract}.csv'
    fig_title_top = f'{date}_{Reaction}_{ML_model}_prop{add_model}_{Search_method}_top{catal_number}_{extract}'
    out_fig_name_top = f'{PATH}/{date}_{Reaction}_{ML_model}_prop{add_model}_{Search_method}_top{catal_number}_{extract}.png'
    out_csv_file_K = f"{PATH}/{date}_{Reaction}_{ML_model}_prop{add_model}_{Search_method}_K{K_cluster}_{extract}.csv"
    fig_title_K = f'{date}_{Reaction}_{ML_model}_prop{add_model}_{Search_method}_K{K_cluster}_{extract}'
    out_fig_name_K = f"{PATH}/{date}_{Reaction}_{ML_model}_prop{add_model}_{Search_method}_K{K_cluster}_{extract}.png"

cand_elem_cols, cand_wt_cols, cand_labels = data_cols['cand_elem'], data_cols['cand_wt'], data_cols['cand_labels']

cand_dict = converter.cand_convert(condition, converted, data_cols, desc_cols,
                                select_desc=True, use_models=[], local=False)
cand, feat, target = cand_dict['cand'], cand_dict['feat'], cand_dict['target']


if pgm_num != 0 and extract in pgm_list:
    print(f'Extract {extract}-containing catalyst candidates in PGMs...')
    if cand_pgm_num == 0:
        print('ERROR: There are no PGMs in the candidate catalysts!')

    elif cand_pgm_num == 1:
        cand_pgm1 = cand[cand['PGM1'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] != 0.0]
        cand = pd.DataFrame()
        cand = cand_pgm1.copy()

    elif cand_pgm_num == 2:
        cand_pgm1 = cand[cand['PGM1'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] != 0.0]

        cand_pgm2 = cand[cand['PGM2'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_pgm2 = cand_pgm2[cand_pgm2['PGM2_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_pgm2 = cand_pgm2[cand_pgm2['PGM2_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_pgm2 = cand_pgm2[cand_pgm2['PGM2_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_pgm2 = cand_pgm2[cand_pgm2['PGM2_wt%'] != 0.0]
        cand = pd.DataFrame()
        cand = pd.concat([cand_pgm1, cand_pgm2])

    elif cand_pgm_num == 3:
        cand_pgm1 = cand[cand['PGM1'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_pgm1 = cand_pgm1[cand_pgm1['PGM1_wt%'] != 0.0]

        cand_pgm2 = cand[cand['PGM2'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_pgm2 = cand_pgm2[cand_pgm2['PGM2_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_pgm2 = cand_pgm2[cand_pgm2['PGM2_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_pgm2 = cand_pgm2[cand_pgm2['PGM2_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_pgm2 = cand_pgm2[cand_pgm2['PGM2_wt%'] != 0]

        cand_pgm3 = cand[cand['PGM3'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_pgm3 = cand_pgm3[cand_pgm3['PGM3_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_pgm3 = cand_pgm3[cand_pgm3['PGM3_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_pgm3 = cand_pgm3[cand_pgm3['PGM3_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_pgm3 = cand_pgm3[cand_pgm3['PGM3_wt%'] != 0.0]
        cand = pd.DataFrame()
        cand = pd.concat([cand_pgm1, cand_pgm2, cand_pgm3])


elif add_num != 0 and extract in add_list:
    print(f'Extract {extract}-containing catalyst candidates in Additivess...')
    if cand_add_num == 0:
        print('ERROR: There are no Additives in the candidate catalysts!')

    elif cand_add_num == 1:
        cand_Ad1 = cand[cand['Ad1'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] != 0.0]
        cand = pd.DataFrame()
        cand = cand_Ad1

    elif cand_add_num == 2:
        cand_Ad1 = cand[cand['Ad1'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] != 0.0]

        cand_Ad2 = cand[cand['Ad2'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] != 0]
        cand = pd.DataFrame()
        cand = pd.concat([cand_Ad1, cand_Ad2])

    elif cand_add_num == 3:
        cand_Ad1 = cand[cand['Ad1'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] != 0.0]

        cand_Ad2 = cand[cand['Ad2'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] != 0]

        cand_Ad3 = cand[cand['Ad3'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] != 0.0]
        cand = pd.DataFrame()
        cand = pd.concat([cand_Ad1, cand_Ad2, cand_Ad3])

    elif cand_add_num == 4:
        cand_Ad1 = cand[cand['Ad1'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] != 0.0]

        cand_Ad2 = cand[cand['Ad2'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] != 0]

        cand_Ad3 = cand[cand['Ad3'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] != 0.0]

        cand_Ad4 = cand[cand['Ad4'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] != 0.0]
        cand = pd.DataFrame()
        cand = pd.concat([cand_Ad1, cand_Ad2, cand_Ad3, cand_Ad4])

    elif cand_add_num == 5:
        cand_Ad1 = cand[cand['Ad1'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] != 0.0]

        cand_Ad2 = cand[cand['Ad2'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] != 0]

        cand_Ad3 = cand[cand['Ad3'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] != 0.0]

        cand_Ad4 = cand[cand['Ad4'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] != 0.0]

        cand_Ad5 = cand[cand['Ad5'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad5 = cand_Ad5[cand_Ad5['Ad5_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad5 = cand_Ad5[cand_Ad5['Ad5_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad5 = cand_Ad5[cand_Ad5['Ad5_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad5 = cand_Ad5[cand_Ad5['Ad5_wt%'] != 0.0]
        cand = pd.DataFrame()
        cand = pd.concat([cand_Ad1, cand_Ad2, cand_Ad3, cand_Ad4, cand_Ad5])

    elif cand_add_num == 6:
        cand_Ad1 = cand[cand['Ad1'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad1 = cand_Ad1[cand_Ad1['Ad1_wt%'] != 0.0]

        cand_Ad2 = cand[cand['Ad2'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad2 = cand_Ad2[cand_Ad2['Ad2_wt%'] != 0]

        cand_Ad3 = cand[cand['Ad3'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad3 = cand_Ad3[cand_Ad3['Ad3_wt%'] != 0.0]

        cand_Ad4 = cand[cand['Ad4'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad4 = cand_Ad4[cand_Ad4['Ad4_wt%'] != 0.0]

        cand_Ad5 = cand[cand['Ad5'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad5 = cand_Ad5[cand_Ad5['Ad5_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad5 = cand_Ad5[cand_Ad5['Ad5_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad5 = cand_Ad5[cand_Ad5['Ad5_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad5 = cand_Ad5[cand_Ad5['Ad5_wt%'] != 0.0]

        cand_Ad6 = cand[cand['Ad6'] == extract]
        if extract_wt > 0.0 and extract_range == 'small':
            cand_Ad6 = cand_Ad6[cand_Ad6['Ad6_wt%'] <= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'large':
            cand_Ad6 = cand_Ad6[cand_Ad6['Ad6_wt%'] >= extract_wt]
        elif extract_wt > 0.0 and extract_range == 'equal':
            cand_Ad6 = cand_Ad6[cand_Ad6['Ad6_wt%'] == extract_wt]
        elif extract_wt == -1:
            cand_Ad6 = cand_Ad6[cand_Ad6['Ad6_wt%'] != 0.0]
        cand = pd.DataFrame()
        cand = pd.concat([cand_Ad1, cand_Ad2, cand_Ad3, cand_Ad4, cand_Ad5, cand_Ad6])


elif supp_num != 0 and extract in supp_list:
    if cand_supp_num == 0:
        print('The support is fixed...')
    elif cand_supp_num == 1:
        print(f'Extract the candidate catalysts containing {extract} as a support...')
        cand = cand[cand['Support'] == extract]

else:
    print('!!!!!!!!!! >>>>>>>>>> ERROR <<<<<<<<<< !!!!!!!!!!')
    print(f'{extract} is not present in the catalyst candidates. Please check for your input errors.')
    print('extract.py is shut down...')
    sys.exit()

cand = converter.cand_str(cand, condition, data_cols)

top_catal = cand.sort_values('ei')
top_catal = top_catal.iloc[-catal_number:, :]
x_top = np.arange(len(top_catal))

if catal_number <= 50:
    plt.figure(figsize=(10,20), facecolor='white')
    plt.barh(x_top,top_catal.loc[:,'ei'], color='blue', alpha=0.7)
    plt.yticks(x_top, top_catal.loc[:, 'Top catal.'], fontsize=14)
    plt.title(fig_title_top, fontsize=14)
    plt.xlabel('Expected improvement (EI)', fontsize=14)
    plt.ylabel('Candidate catalyst composition', fontsize=14)
    plt.savefig(out_fig_name_top, dpi=600, bbox_inches="tight")

elif catal_number <= 100 and catal_number > 50:
    fig = plt.figure(figsize=(20,20), facecolor='white')
    ax1 = fig.add_subplot(1,2,1)
    ax2 = fig.add_subplot(1,2,2)
    fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=1)
    
    top_catal1 = top_catal.iloc[0:int(1/2*catal_number),:]
    top_catal2 = top_catal.iloc[int(1/2*catal_number):catal_number,:]
    x_top1 = np.arange(int(1/2*len(top_catal)))

    fig.suptitle(fig_title_top, x=0.35, y=1.03, fontsize=14)
    ax1.barh(x_top1, top_catal1.loc[:,'ei'], color='blue', alpha=0.7)
    ax1.invert_yaxis()
    ax1.set_xlabel('Expected improvement (EI)', fontsize=14)
    ax1.set_yticks(x_top1, fontsize=14)
    ax1.set_ylabel('Candidate catalyst composition', fontsize=14)
    ax1.set_yticklabels(top_catal1.loc[:, 'Top catal.'], fontsize=14)

    ax2.barh(x_top1, top_catal2.loc[:,'ei'], color='blue', alpha=0.7)
    ax2.invert_yaxis()
    ax2.set_xlabel('Expected improvement (EI)', fontsize=14)
    ax2.set_yticks(x_top1, fontsize=14)
    ax2.set_ylabel('Candidate catalyst composition', fontsize=14)
    ax2.set_yticklabels(top_catal2.loc[:, 'Top catal.'], fontsize=14)
    plt.savefig(out_fig_name_top, dpi=600, bbox_inches="tight")

elif catal_number <= 150 and catal_number > 100:
    fig = plt.figure(figsize=(30,20), facecolor='white')
    ax1 = fig.add_subplot(1,3,1)
    ax2 = fig.add_subplot(1,3,2)
    ax3 = fig.add_subplot(1,3,3)
    fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=1)
    
    top_catal1 = top_catal.iloc[0:int(1/3*catal_number),:]
    top_catal2 = top_catal.iloc[int(1/3*catal_number):int(2/3*catal_number),:]
    top_catal3 = top_catal.iloc[int(2/3*catal_number):catal_number,:]
    x_top1 = np.arange(int(1/3*len(top_catal)))

    fig.suptitle(fig_title_top, x=0.4, y=1.03, fontsize=14)
    ax1.barh(x_top1, top_catal1.loc[:,'ei'], color='blue', alpha=0.7)
    ax1.invert_yaxis()
    ax1.set_xlabel('Expected improvement (EI)', fontsize=14)
    ax1.set_yticks(x_top1, fontsize=14)
    ax1.set_ylabel('Candidate catalyst composition', fontsize=14)
    ax1.set_yticklabels(top_catal1.loc[:, 'Top catal.'], fontsize=14)

    ax2.barh(x_top1, top_catal2.loc[:,'ei'], color='blue', alpha=0.7)
    ax2.invert_yaxis()
    ax2.set_xlabel('Expected improvement (EI)', fontsize=14)
    ax2.set_yticks(x_top1, fontsize=14)
    ax2.set_ylabel('Candidate catalyst composition', fontsize=14)
    ax2.set_yticklabels(top_catal2.loc[:, 'Top catal.'], fontsize=14)

    ax3.barh(x_top1, top_catal3.loc[:,'ei'], color='blue', alpha=0.7)
    ax3.invert_yaxis()
    ax3.set_xlabel('Expected improvement (EI)', fontsize=14)
    ax3.set_yticks(x_top1, fontsize=14)
    ax3.set_ylabel('Candidate catalyst composition', fontsize=14)
    ax3.set_yticklabels(top_catal3.loc[:, 'Top catal.'], fontsize=14)
    plt.savefig(out_fig_name_top, dpi=600, bbox_inches="tight")

elif catal_number > 150:
    fig = plt.figure(figsize=(40,20), facecolor='white')
    ax1 = fig.add_subplot(1,4,1)
    ax2 = fig.add_subplot(1,4,2)
    ax3 = fig.add_subplot(1,4,3)
    ax4 = fig.add_subplot(1,4,4)
    fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=1)
    
    top_catal1 = top_catal.iloc[0:int(1/4*catal_number),:]
    top_catal2 = top_catal.iloc[int(1/4*catal_number):int(1/2*catal_number),:]
    top_catal3 = top_catal.iloc[int(1/2*catal_number):int(3/4*catal_number),:]
    top_catal4 = top_catal.iloc[int(3/4*catal_number):catal_number,:]
    x_top1 = np.arange(int(1/4*len(top_catal)))

    fig.suptitle(fig_title_top, x=0.43, y=1.03, fontsize=14)
    ax1.barh(x_top1, top_catal1.loc[:,'ei'], color='blue', alpha=0.7)
    ax1.invert_yaxis()
    ax1.set_xlabel('Expected improvement (EI)', fontsize=14)
    ax1.set_yticks(x_top1, fontsize=14)
    ax1.set_ylabel('Candidate catalyst composition', fontsize=14)
    ax1.set_yticklabels(top_catal1.loc[:, 'Top catal.'], fontsize=14)

    ax2.barh(x_top1, top_catal2.loc[:,'ei'], color='blue', alpha=0.7)
    ax2.invert_yaxis()
    ax2.set_xlabel('Expected improvement (EI)', fontsize=14)
    ax2.set_yticks(x_top1, fontsize=14)
    ax2.set_ylabel('Candidate catalyst composition', fontsize=14)
    ax2.set_yticklabels(top_catal2.loc[:, 'Top catal.'], fontsize=14)

    ax3.barh(x_top1, top_catal3.loc[:,'ei'], color='blue', alpha=0.7)
    ax3.invert_yaxis()
    ax3.set_xlabel('Expected improvement (EI)', fontsize=14)
    ax3.set_yticks(x_top1, fontsize=14)
    ax3.set_ylabel('Candidate catalyst composition', fontsize=14)
    ax3.set_yticklabels(top_catal3.loc[:, 'Top catal.'], fontsize=14)

    ax4.barh(x_top1, top_catal4.loc[:,'ei'], color='blue', alpha=0.7)
    ax4.invert_yaxis()
    ax4.set_xlabel('Expected improvement (EI)', fontsize=14)
    ax4.set_yticks(x_top1, fontsize=14)
    ax4.set_ylabel('Candidate catalyst composition', fontsize=14)
    ax4.set_yticklabels(top_catal4.loc[:, 'Top catal.'], fontsize=14)
    plt.savefig(out_fig_name_top, dpi=600, bbox_inches="tight")

if len(cand) > K_cluster:
    cvmodel, error = analysis.grid_search_CV(feat, target, split_type='KFold', split_num=10, use_model='ETR')
    print('n_estimators=', cvmodel.best_params_['n_estimators'], 'bootstrap=', cvmodel.best_params_['bootstrap'])
    model  = ExtraTreesRegressor(n_estimators=cvmodel.best_params_['n_estimators'], bootstrap=cvmodel.best_params_['bootstrap'], random_state=1107, n_jobs=-1)
    model.fit(feat, target)
    k = KMeans(n_clusters=K_cluster, random_state=1107)
    cluster = k.fit_predict(cand.loc[:, cand_wt_cols])
    cluster = pd.Series(cluster, index=cand.index, name='cluster')

    cand_top = cand.copy()
    cand = pd.concat([cand, cluster], axis=1)
    cand = cand.sort_values(by=['cluster','ei']).drop_duplicates(subset=['cluster'],keep='last')
    cand = cand.sort_values(by='ei', ascending=False)
    cand.to_csv(out_csv_file_K)
    print(f'cand_K{K_cluster}_shape:', cand.shape)
    cand = converter.cand_str(cand, cand_pgm_num, cand_add_num, cand_supp_num, cand_wt_cols)

    K_catal = cand.sort_values('ei')
    K_catal = cand.iloc[-K_cluster:, :]
    x_K = np.arange(len(K_catal))

    if K_cluster <= 50:
        plt.figure(figsize=(10,20), facecolor='white')
        plt.yticks(x_K, K_catal.loc[:, 'Top catal.'], fontsize=14)
        plt.barh(x_K, K_catal.loc[:,'ei'], color='blue', alpha=0.7)
        plt.gca().invert_yaxis()
        plt.title(fig_title_K, fontsize=14)
        plt.xlabel('Expected improvement (EI)', fontsize=14)
        plt.ylabel('Candidate catalyst composition', fontsize=14)
        plt.savefig(out_fig_name_K, dpi=600, bbox_inches="tight")

    elif K_cluster <= 100 and K_cluster > 50:
        fig = plt.figure(figsize=(20,20), facecolor='white')
        ax1 = fig.add_subplot(1,2,1)
        ax2 = fig.add_subplot(1,2,2)
        fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=1)
        
        K_catal1 = K_catal.iloc[0:int(1/2*K_cluster),:]
        K_catal2 = K_catal.iloc[int(1/2*K_cluster):K_cluster,:]
        x_K1 = np.arange(int(1/2*len(K_catal)))

        fig.suptitle(fig_title_K, x=0.35, y=1.03, fontsize=14)
        ax1.barh(x_K1, K_catal1.loc[:,'ei'], color='blue', alpha=0.7)
        ax1.invert_yaxis()
        ax1.set_xlabel('Expected improvement (EI)', fontsize=14)
        ax1.set_yticks(x_K1, fontsize=14)
        ax1.set_ylabel('Candidate catalyst composition', fontsize=14)
        ax1.set_yticklabels(K_catal1.loc[:, 'Top catal.'], fontsize=14)

        ax2.barh(x_K1, K_catal2.loc[:,'ei'], color='blue', alpha=0.7)
        ax2.invert_yaxis()
        ax2.set_xlabel('Expected improvement (EI)', fontsize=14)
        ax2.set_yticks(x_K1, fontsize=14)
        ax2.set_ylabel('Candidate catalyst composition', fontsize=14)
        ax2.set_yticklabels(K_catal2.loc[:, 'Top catal.'], fontsize=14)
        plt.savefig(out_fig_name_K, dpi = 600, bbox_inches = "tight")
        
    elif K_cluster <= 150 and K_cluster > 100:
        fig = plt.figure(figsize=(30,20), facecolor='white')
        ax1 = fig.add_subplot(1,3,1)
        ax2 = fig.add_subplot(1,3,2)
        ax3 = fig.add_subplot(1,3,3)
        fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=1)
        
        K_catal1 = K_catal.iloc[0:int(1/3*K_cluster),:]
        K_catal2 = K_catal.iloc[int(1/3*K_cluster):int(2/3*K_cluster),:]
        K_catal3 = K_catal.iloc[int(2/3*K_cluster):K_cluster,:]
        x_K1 = np.arange(int(1/3*len(K_catal)))

        fig.suptitle(fig_title_K, x=0.4, y=1.03, fontsize=14)
        ax1.barh(x_K1, K_catal1.loc[:,'ei'], color='blue', alpha=0.7)
        ax1.invert_yaxis()
        ax1.set_xlabel('Expected improvement (EI)', fontsize=14)
        ax1.set_yticks(x_K1, fontsize=14)
        ax1.set_ylabel('Candidate catalyst composition', fontsize=14)
        ax1.set_yticklabels(K_catal1.loc[:, 'Top catal.'], fontsize=14)

        ax2.barh(x_K1, K_catal2.loc[:,'ei'], color='blue', alpha=0.7)
        ax2.invert_yaxis()
        ax2.set_xlabel('Expected improvement (EI)', fontsize=14)
        ax2.set_yticks(x_K1, fontsize=14)
        ax2.set_ylabel('Candidate catalyst composition', fontsize=14)
        ax2.set_yticklabels(K_catal2.loc[:, 'Top catal.'], fontsize=14)

        ax3.barh(x_K1, K_catal3.loc[:,'ei'], color='blue', alpha=0.7)
        ax3.invert_yaxis()
        ax3.set_xlabel('Expected improvement (EI)', fontsize=14)
        ax3.set_yticks(x_K1, fontsize=14)
        ax3.set_ylabel('Candidate catalyst composition', fontsize=14)
        ax3.set_yticklabels(K_catal3.loc[:, 'Top catal.'], fontsize=14)
        plt.savefig(out_fig_name_K, dpi = 600, bbox_inches = "tight")

    elif K_cluster > 150:
        fig = plt.figure(figsize=(40,20), facecolor='white')
        ax1 = fig.add_subplot(1,4,1)
        ax2 = fig.add_subplot(1,4,2)
        ax3 = fig.add_subplot(1,4,3)
        ax4 = fig.add_subplot(1,4,4)
        fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=1)
        
        K_catal1 = K_catal.iloc[0:int(1/4*K_cluster),:]
        K_catal2 = K_catal.iloc[int(1/4*K_cluster):int(1/2*K_cluster),:]
        K_catal3 = K_catal.iloc[int(1/2*K_cluster):int(3/4*K_cluster),:]
        K_catal4 = K_catal.iloc[int(3/4*K_cluster):K_cluster,:]
        x_K1 = np.arange(int(1/4*len(K_catal)))

        fig.suptitle(fig_title_K, x=0.43, y=1.03, fontsize=14)
        ax1.barh(x_K1, K_catal1.loc[:,'ei'], color='blue', alpha=0.7)
        ax1.invert_yaxis()
        ax1.set_xlabel('Expected improvement (EI)', fontsize=14)
        ax1.set_yticks(x_K1, fontsize=14)
        ax1.set_ylabel('Candidate catalyst composition', fontsize=14)
        ax1.set_yticklabels(K_catal1.loc[:, 'Top catal.'], fontsize=14)

        ax2.barh(x_K1, K_catal2.loc[:,'ei'], color='blue', alpha=0.7)
        ax2.invert_yaxis()
        ax2.set_xlabel('Expected improvement (EI)', fontsize=14)
        ax2.set_yticks(x_K1, fontsize=14)
        ax2.set_ylabel('Candidate catalyst composition', fontsize=14)
        ax2.set_yticklabels(K_catal2.loc[:, 'Top catal.'], fontsize=14)

        ax3.barh(x_K1, K_catal3.loc[:,'ei'], color='blue', alpha=0.7)
        ax3.invert_yaxis()
        ax3.set_xlabel('Expected improvement (EI)', fontsize=14)
        ax3.set_yticks(x_K1, fontsize=14)
        ax3.set_ylabel('Candidate catalyst composition', fontsize=14)
        ax3.set_yticklabels(K_catal3.loc[:, 'Top catal.'], fontsize=14)

        ax4.barh(x_K1, K_catal4.loc[:,'ei'], color='blue', alpha=0.7)
        ax4.invert_yaxis()
        ax4.set_xlabel('Expected improvement (EI)', fontsize=14)
        ax4.set_yticks(x_K1, fontsize=14)
        ax4.set_ylabel('Candidate catalyst composition', fontsize=14)
        ax4.set_yticklabels(K_catal4.loc[:, 'Top catal.'], fontsize=14)
        plt.savefig(out_fig_name_K, dpi = 600, bbox_inches = "tight")

elif len(cand) <= K_cluster:
    print(f'The number of extracted catalyst candidates is less than {K_cluster}, so clustering was not possible.')
