#!/usr/bin/env python
import os, sys, glob
import time
import gc
import subprocess
import os.path
sys.dont_write_bytecode = True

from concurrent.futures import ProcessPoolExecutor

import numpy as np
import pandas as pd
import scipy.sparse as sparse
from sklearn.ensemble import ExtraTreesRegressor

import modules.conditions as conditions
import modules.setting as setting
import modules.converter as converter

import random
random.seed(1107)
np.random.seed(1107)

import argparse
parser = argparse.ArgumentParser()
parser.add_argument('--num',  type=int, default=0)
parser.add_argument('--cores', type=int, default=36)
parser.add_argument('--from', type=int, default=0)
parser.add_argument('--to', type=int, default=1000)

args = vars(parser.parse_args())
job_num = args['num']
num_cores = args['cores']
idx_from = args['from']
idx_to = args['to']

########################################################################################################
########################################################################################################

condition = conditions.calc_condition()
computer = condition['computer']
date, Reaction, data_sheet_name = condition['date'], condition['Reaction'], condition['data_sheet_name']
cut_size, save_depth = condition['cut_size'], condition['save_depth']

pgm_model = condition['pgm_model']
add_model = condition['add_model']
supp_model = condition['supp_model']

cand_pgm_num = condition['cand_pgm_num']
cand_add_num = condition['cand_add_num']
cand_supp_num = condition['cand_supp_num']

data_cols = setting.data_columns(condition)
desc_cols = setting.desc_columns(Reaction)
converted = converter.data_convert(condition, data_cols, desc_cols,
									sheet_name='data_sheet_name', select_desc=True,
                                    use_models=[], local=False, idx=None)

feat, target = converted['feat'], converted['target']
cur_max, cur_min = target.max(), target.min()

elem_lists = converter.elem_list(condition, converted, desc_cols)
cand_cols_order = converter.cols_order(condition)

# print('# cpu cores', os.cpu_count())
# print('num_cores', num_cores)
print(date, Reaction, 'pgm_model=', pgm_model, 'add_model=', add_model, 'supp_model=', supp_model)


"""
cvmodel, error = analysis.grid_search_CV(feat, target, split_type='K',
                                         split_num=10, use_model='ETR', plot=False)
model = cvmodel.best_estimator_
print('Best model')
print(cvmodel.best_params_)
"""

model = ExtraTreesRegressor(n_estimators=condition['n_est'], random_state=1107, n_jobs=-1)
model.fit(np.array(feat), np.array(target))


cand_elem_wt, num_elem_wt = converter.all_wt_comb(condition)
total_elem_num = cand_pgm_num + cand_add_num + cand_supp_num
cand_pgm_wt = np.array(cand_elem_wt).reshape(-1, total_elem_num)[:, 0: cand_pgm_num]
cand_add_wt = np.array(cand_elem_wt).reshape(-1, total_elem_num)[:, cand_pgm_num: cand_pgm_num + cand_add_num]

if cand_supp_num != 0:
	cand_supp_wt = np.array(cand_elem_wt).reshape(-1, total_elem_num)[:, -cand_supp_num]
else:
	cand_supp_wt = []

part_elem_wt = cand_elem_wt[idx_from:idx_to]
part_pgm_wt = cand_pgm_wt[idx_from:idx_to]
part_add_wt = cand_add_wt[idx_from:idx_to]


all_comb, total_comb_num = converter.catal_comb(condition, elem_lists)
divisors = converter.divisors_list(total_comb_num)
slice_size, num_slice = converter.num_jobs(total_comb_num, divisors, size=cut_size)
print('total_wt_num:', num_elem_wt)
print('total_comb_num:', total_comb_num)
print('Slice size:', slice_size)
print('Slice (Cycle) num:', num_slice)

# Test
start, end = 0, slice_size
all_comb_slice, pgm_rep, pgm_oh, add_rep, add_oh, supp_rep, supp_oh \
		= converter.input_comps(condition, converted, elem_lists, cand_cols_order, slice_size, start, end)
print('Test OK !')


########################################################################################################
########################################################################################################

def calc_ei(cand_elem_wt: list, cand_pgm_wt: list, cand_add_wt: list) -> pd.DataFrame:
    """
    Calculate the Expected Improvement (EI) value for a given catalyst composition based on learned models.

    This function computes a weighted feature representation (elem_pattern) for a catalyst composition
    by combining the contributions from platinum group metals (PGM), additives (Add), and support materials (Supp).
    The weighting is determined by the candidate weight lists (cand_pgm_wt, cand_add_wt, and implicitly cand_supp_wt)
    and the corresponding one-hot (e.g., pgm_oh, add_oh, supp_oh) or descriptor (e.g., pgm_rep, add_rep, supp_rep)
    representations. The combination strategy depends on the model flags (pgm_model, add_model, supp_model) and the 
    candidate numbers (cand_pgm_num, cand_add_num, cand_supp_num).

    The ensemble model (global variable `model`) is then used to predict the mean (mu) for the input pattern,
    and the standard deviation (sigma) is computed using `converter.pred_std_csr`. To avoid division-by-zero 
    issues, any zero sigma values are replaced with a small constant (1e-5). Finally, the Expected Improvement 
    (EI) score is computed using `converter.EI`, and the candidate compositions are ranked based on their EI values.
    The function returns a DataFrame with the EI scores of the top candidates (as determined by the global 
    variable `save_depth`), with the index provided by `converter.get_encode`.

    Note: This function relies on several global variables that must be defined elsewhere in the code:
          - pgm_model, cand_pgm_num, pgm_oh, pgm_rep
          - add_model, cand_add_num, add_oh, add_rep
          - supp_model, cand_supp_num, supp_oh, supp_rep, cand_supp_wt
          - model (a trained ensemble model with a predict method and an estimators_ attribute)
          - Reaction, cur_min, cur_max, save_depth
          - cand_elem_wt, all_comb_slice, condition
          - converter (an object or module providing pred_std_csr, EI, and get_encode functions)
          
    Parameters
    ----------
    cand_elem_wt : list
        Candidate element weight percentages (used for encoding the candidate compositions).
    cand_pgm_wt : list
        Weight percentages for the PGM components of the candidate catalyst.
    cand_add_wt : list
        Weight percentages for the additive components of the candidate catalyst.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the EI scores for the selected candidate catalyst compositions, with the index
        provided by the encoded catalyst composition strings.
    """
    # Compute PGM feature pattern.
    if pgm_model == 0 and cand_pgm_num != 0:
        pgm_pattern = sum([d * w for d, w in zip(pgm_oh, cand_pgm_wt)])
    elif pgm_model == 1 and cand_pgm_num != 0:
        pgm_rep_pattern = sum([d * w for d, w in zip(pgm_rep, cand_pgm_wt)])
        pgm_oh_pattern = sum([d * w for d, w in zip(pgm_oh, cand_pgm_wt)])
        pgm_pattern = sparse.hstack([pgm_rep_pattern, pgm_oh_pattern])
        del pgm_rep_pattern
        del pgm_oh_pattern
    elif pgm_model == 2 and cand_pgm_num != 0:
        pgm_pattern = sum([d * w for d, w in zip(pgm_rep, cand_pgm_wt)])
    else:
        pgm_pattern = None

    # Compute Additive feature pattern.
    if add_model == 0 and cand_add_num != 0:
        add_pattern = sum([d * w for d, w in zip(add_oh, cand_add_wt)])
    elif add_model == 1 and cand_add_num != 0:
        add_rep_pattern = sum([d * w for d, w in zip(add_rep, cand_add_wt)])
        add_oh_pattern = sum([d * w for d, w in zip(add_oh, cand_add_wt)])
        add_pattern = sparse.hstack([add_rep_pattern, add_oh_pattern])
        del add_rep_pattern
        del add_oh_pattern
    elif add_model == 2 and cand_add_num != 0:
        add_pattern = sum([d * w for d, w in zip(add_rep, cand_add_wt)])
    else:
        add_pattern = None

    # Compute Support material feature pattern.
    if supp_model == 0 and cand_supp_num != 0:
        supp_pattern = sum([d * w for d, w in zip(supp_oh, cand_supp_wt)])
    elif supp_model == 1 and cand_supp_num != 0:
        supp_rep_pattern = sum([d * w for d, w in zip(supp_rep, cand_supp_wt)])
        supp_oh_pattern = sum([d * w for d, w in zip(supp_oh, cand_supp_wt)])
        supp_pattern = sparse.hstack([supp_rep_pattern, supp_oh_pattern])
        del supp_rep_pattern
        del supp_oh_pattern
    elif supp_model == 2 and cand_supp_num != 0:
        supp_pattern = sum([d * w for d, w in zip(supp_rep, cand_supp_wt)])
    else:
        supp_pattern = None

    # Combine feature patterns from all components.
    pattern_list = [pgm_pattern, add_pattern, supp_pattern]
    pattern_list = [i for i in pattern_list if i is not None]

    if len(pattern_list) > 1:
        elem_pattern = sparse.hstack(pattern_list)
    else:
        elem_pattern = pattern_list[0]

    # Predict the mean and compute the standard deviation.
    mu = model.predict(elem_pattern)
    sigma = converter.pred_std_csr(elem_pattern, model.estimators_, mu)
    idx_s = sigma == 0
    sigma[idx_s] = 1e-5

    # Calculate Expected Improvement (EI) score.
    scores = converter.EI(Reaction, mu, sigma, cur_min, cur_max)
    if save_depth == -1:
        idx = scores.argsort()[::-1]
    else:
        idx = scores.argsort()[::-1][:save_depth]

    # Return a DataFrame with EI scores for the selected candidates, using encoded catalyst composition strings as index.
    return pd.DataFrame({'ei': scores[idx]}, index=converter.get_encode(idx, cand_elem_wt, all_comb_slice, condition))

########################################################################################################
########################################################################################################


print(f'EI calculation start ({num_slice} cycle).')
os.makedirs(f'cand/cand_{job_num}', exist_ok=True)
for i, j in enumerate(zip(np.arange(0, total_comb_num, slice_size), np.arange(slice_size, total_comb_num+slice_size, slice_size))):
	print(f'---------- Cycle {i}/{num_slice} start... ----------')
	start_time = time.perf_counter()
	comb_slice_list, pgm_rep, pgm_oh, add_rep, add_oh, supp_rep, supp_oh \
         = converter.input_comps(condition, converted, elem_lists, cand_cols_order, slice_size, j[0], j[1])

	with ProcessPoolExecutor(max_workers=num_cores) as executor:
		res = executor.map(calc_ei, part_elem_wt, part_pgm_wt, part_add_wt)
	cand = pd.concat(res)
	cand.to_csv(f'cand/cand_{job_num}/cand_{job_num}_{i}.csv')

	del pgm_rep, pgm_oh, add_rep, add_oh, supp_rep, supp_oh
	del res, cand

	gc.collect()
	end_time = time.perf_counter()
	elapsed_time = (end_time - start_time) / 60
	print('TIME:', round(elapsed_time, 2), '(min)')
	print(f'---------- Cycle {i}/{num_slice} finished. ----------')

csv_files = glob.glob(f'cand/cand_{job_num}/*.csv')
data_list = []
for file in csv_files:
	print(file)
	data_list.append(pd.read_csv(file))

cand = pd.concat(data_list, axis=0, sort=True)
cand.to_csv(f'cand/cand_sum_{job_num}.csv', index=False)
print('All EI calculations were completed !!')

if idx_to == num_elem_wt:
	K_cluster = condition['K_cluster']
	subprocess.call(f"python sum_k_clus_{Reaction}.py -k {K_cluster}", shell=True)
	print('All calculations and K-cluster calculations have been successfully completed !!')
	
	extractions = condition['extractions']
	if len(extractions) != 0:
		for i in range(len(extractions)):
			subprocess.call(f"python extract_{Reaction}.py -ex {extractions[i]}", shell=True)
			print(f'K-clustering was attempted by extracting candidate catalysts that contained {extractions[i]}.')
	else:
		print('Extractions were NOT performed...')
	print('All calculations have been successfully completed !!')
