#!/usr/bin/env python
import os, sys
sys.dont_write_bytecode = True

import numpy as np
import pandas as pd
import collections
import math
import itertools
import matplotlib as mpl
import matplotlib.pyplot as plt

from sklearn.ensemble import ExtraTreesRegressor, RandomForestRegressor
from sklearn.linear_model import Lasso, Ridge
from sklearn.svm import SVR
import sklearn.gaussian_process as gp
from sklearn.model_selection import ShuffleSplit, KFold, GridSearchCV, train_test_split
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score
from sklearn.model_selection import train_test_split, LeaveOneOut

from adjustText import adjust_text
import shap

import img2pdf
from PIL import Image

#import modules.conditions as conditions
import modules.conditions_CH3OH_Batch as conditions

import modules.converter as converter

import random
random.seed(1107)
np.random.seed(1107)

### Specify figure resolution ###
dpi_value = 150

plt.rcParams['font.size'] = 18
plt.rcParams['xtick.labelsize'] = 17
plt.rcParams['ytick.labelsize'] = 17
plt.rcParams['font.family'] = 'DejaVu Sans'
plt.rcParams['font.sans-serif'] = ['Arial']
plt.rcParams["mathtext.fontset"] = "stixsans"
plt.rcParams.update({'mathtext.default':'regular'})
plt.rcParams['xtick.direction'] = 'out'
plt.rcParams['ytick.direction'] = 'out'
plt.rcParams['xtick.major.width'] = 1.0
plt.rcParams['ytick.major.width'] = 1.0
plt.rcParams['axes.linewidth'] = 1.0
plt.rcParams['axes.grid'] = False
plt.rcParams['grid.linestyle'] = '--'
plt.rcParams['grid.linewidth'] = 2
plt.rcParams["legend.markerscale"] = 2
plt.rcParams["legend.fancybox"] = False
plt.rcParams["legend.framealpha"] = 1
plt.rcParams["legend.edgecolor"] = 'black'


########################################################################################################
########################################################################################################

condition = conditions.calc_condition()
date, Reaction, data_sheet_name = condition['date'], condition['Reaction'], condition['data_sheet_name']
target_name = condition['target_name']

if Reaction == 'CH3OH_Batch':
    target_short_name = '$STY_{CH_{3}OH}$'
    y_label_name = '$CH_{3}OH$ formation rate (mmol $g^{-1}$ $h^{-1}$)'


DIR_svg = f'{date}_{Reaction}_EDA_svg'
DIR_csv = f'{date}_{Reaction}_CSV_file'
DIR_pdf = f'{date}_{Reaction}_EDA_pdf'
os.makedirs(DIR_svg, exist_ok=True)
os.makedirs(DIR_csv, exist_ok=True)
os.makedirs(DIR_pdf, exist_ok=True)

########################################################################################################
########################################################################################################


# for violin plot (1)
def set_axis_style(ax: mpl.axes.Axes, labels: list):
    """
    Set the style of the x-axis for a given axis object.

    This function customizes the x-axis of a matplotlib axis object, including tick direction, 
    tick position, tick labels, and axis limits.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The matplotlib axis object to customize.
    labels : list
        A list of labels for the x-axis ticks.

    Returns
    -------
    None

    Notes
    -----
    - The x-axis ticks are set to point outward (`direction='out'`).
    - The tick position is fixed at the bottom of the axis.
    - The x-axis limits are adjusted to provide padding around the ticks.
    """
    ax.xaxis.set_tick_params(direction='out')
    ax.xaxis.set_ticks_position('bottom')
    ax.set_xticks(np.arange(1, len(labels) + 1), labels=labels)
    ax.set_xlim(0.25, len(labels) + 0.75)



# for violin plot (2)
def adjacent_values(vals: np.array, q1: float, q3: float) -> tuple:
    """
    Calculate the adjacent values for a box plot.

    This function computes the lower and upper adjacent values based on the interquartile range (IQR).
    Adjacent values are clipped to the range of the data.

    Parameters
    ----------
    vals : array-like
        A sorted array of data values.
    q1 : float
        The first quartile (25th percentile) of the data.
    q3 : float
        The third quartile (75th percentile) of the data.

    Returns
    -------
    tuple
        A tuple containing:
        - lower_adjacent_value : float
            The lower adjacent value, clipped to the minimum value of `vals`.
        - upper_adjacent_value : float
            The upper adjacent value, clipped to the maximum value of `vals`.

    Notes
    -----
    - The adjacent values are calculated using the formula:
      - `upper_adjacent_value = q3 + 1.5 * IQR`
      - `lower_adjacent_value = q1 - 1.5 * IQR`
      where `IQR = q3 - q1`.
    - The adjacent values are clipped to ensure they do not exceed the data range (`vals[0]` to `vals[-1]`).
    """
    upper_adjacent_value = q3 + (q3 - q1) * 1.5
    upper_adjacent_value = np.clip(upper_adjacent_value, q3, vals[-1])

    lower_adjacent_value = q1 - (q3 - q1) * 1.5
    lower_adjacent_value = np.clip(lower_adjacent_value, vals[0], q1)
    return lower_adjacent_value, upper_adjacent_value



# Function that sorts a specified range of columns in a given row in descending order
# and formats the column names and values accordingly.
def df_sort(data: pd.DataFrame, data_cols: dict, index_num: int,
            elem_num: int, start_num: int, end_num: int) -> tuple:
    """
    Sort a specified range of columns in the given row in descending order and format the column names and values.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame containing the data.
    data_cols : dict
        Dictionary containing column names, with key 'wt' for wt columns.
    index_num : int
        The index (row) number to extract from the data.
    elem_num : int
        The number of elements to process; used to determine how many sorted columns to extract.
    start_num : int
        The starting index of the column range in wt_cols.
    end_num : int
        The ending index of the column range in wt_cols.

    Returns
    -------
    df_sort_elem : pd.DataFrame
        The DataFrame containing the formatted sorted elements.
    df_sort_wt : pd.DataFrame
        The DataFrame containing the sorted wt values.
    """
    wt_cols = data_cols['wt']
    df_sort_elem = data.loc[index_num][[                    # Extract the row at index_num from data
        data.loc[index_num][wt_cols[start_num : end_num]]   # Obtain a specific range of column names from the wt_cols list
                                                            # --> This part extracts only the columns corresponding to wt_cols from the row at index_num
        .sort_values(ascending=False)                       # Since ascending=False, the extracted data is sorted in descending order by value
        .index[j]                                           # Retrieve the j-th index (column name) from the sorted data (j is used within the loop)
        .replace('_wt%', '')                                # Remove '_wt%' from the obtained column name (index)
        for j in range(elem_num)                            # Loop over the range of elem_num, repeating the above process multiple times
                                                            # "data.loc[index_num][[ ... ]]" --> Using the generated list of column names,
                                                            # extract the corresponding columns again from the row at index_num
        ]].reset_index().T.drop('index', axis=0)            # Reset the index, add a new index, transpose (swap rows and columns), and remove the old index
    
    df_sort_wt = data.loc[index_num]\
        [wt_cols[start_num : end_num]]\
        .sort_values(ascending=False)\
            .reset_index().T.drop('index', axis=0)

    return df_sort_elem, df_sort_wt



def catalyst_name(condition: dict, converted: dict, data_cols: dict) -> pd.DataFrame:

    pgm_num = condition['pgm_num']
    add_num = condition['add_num']
    data = converted['data']
    elem_cols = data_cols['elem']
    wt_cols = data_cols['wt']

    data_sort = pd.DataFrame(index=np.arange(0, len(data), 1),
                             columns=elem_cols + wt_cols)

    for i in range(len(data)):
        df_sort_PGM, df_sort_PGM_wt = \
            df_sort(data=data, data_cols=data_cols, index_num=i, elem_num=pgm_num,
                    start_num=0, end_num=pgm_num)
        
        df_sort_Add, df_sort_Add_wt = \
            df_sort(data=data, data_cols=data_cols, index_num=i, elem_num=add_num,
                    start_num=pgm_num, end_num=pgm_num + add_num)
        
        data_sort_elem = pd.concat([df_sort_PGM, df_sort_Add], axis=1)
        data_sort_wt = pd.concat([df_sort_PGM_wt, df_sort_Add_wt], axis=1)

        temp_df = pd.concat([data_sort_elem, data_sort_wt], axis=1)
        data_sort.loc[i] = temp_df.values

    data_sort = pd.concat([
        data_sort,
        data[['Support_name']],
        data[['Calc T']]
        ], axis=1)

    data_sort['Catalyst'] = ''
    for i in range(len(data)):
        catalyst_name = ''  
        # PGM elements part.
        for j in range(pgm_num):
            if data_sort[elem_cols[j]][i] != 'H':
                if j == 0:
                    catalyst_name += data_sort[elem_cols[j]][i]
                else:
                    catalyst_name += '-' + data_sort[elem_cols[j]][i]

                if data_sort[wt_cols[j]][i] > 0:
                    catalyst_name += f"({data_sort[wt_cols[j]][i]})"
        
        catalyst_name += '/'

        # Additive elements part.
        for j in range(pgm_num, pgm_num + add_num):
            if data_sort[elem_cols[j]][i] != 'H':
                if j == pgm_num:
                    catalyst_name += data_sort[elem_cols[j]][i]
                else:
                    catalyst_name += '-' + data_sort[elem_cols[j]][i]

                if data_sort[wt_cols[j]][i] > 0:
                    catalyst_name += f"({data_sort[wt_cols[j]][i]})"
        
        catalyst_name += '/'

        catalyst_name += f"/{data_sort['Support_name'][i]}"
        catalyst_name += f", CalcT[{data_sort['Calc T'][i]}(℃)]"

        # Replace “//” with “/” 
        catalyst_name = catalyst_name.replace('//', '/')

        # Replace “/-” with “/”  
        catalyst_name = catalyst_name.replace('/-', '/')

         # If the string starts with “/”, remove only the leading “/”   
        catalyst_name = catalyst_name.lstrip("/") if catalyst_name.startswith("/") else catalyst_name

        # If the string starts with “-”,remove only the leading “-
        catalyst_name = catalyst_name.lstrip("-") if catalyst_name.startswith("-") else catalyst_name

        data_sort.at[i, 'Catalyst'] = catalyst_name

    return data_sort



def dup_catal_name_list(data_sort: pd.DataFrame) -> list:
    """
    Detect and display duplicate entries in the 'Catalyst' column of the DataFrame and return the list of duplicates.

    Parameters
    ----------
    data_sort : pd.DataFrame
        The DataFrame containing the 'Catalyst' column to check for duplicates.

    Returns
    -------
    list
        A list of duplicated catalyst names.

    Notes
    -----
    - The function checks for duplicate entries in the 'Catalyst' column and prints the indices of the duplicated rows.
    - If no duplicates are found, it prints a message indicating that there are no duplicates.
    """
    # List the duplicate entries in the 'Catalyst' column
    dup_cat_list = [k for k, v in collections.Counter(data_sort['Catalyst']).items() if v > 1]
    
    if dup_cat_list:
        print('#'*100)
        for i in dup_cat_list:
            # Obtain the duplicated row indices (+1)
            dup_num = list(data_sort[data_sort['Catalyst'] == i].index + 1)
            print(f'No. {dup_num} catalysts were duplicated !!')
            print(i)
            print('')
        print('#'*100)
    else:
        print('There were no duplicates.')
    
    return dup_cat_list



def df_sort_max(condition: dict, converted: dict, data_sort: pd.DataFrame) -> tuple:
    """
    Sort and extract maximum target values from experimental data.

    This function merges sorted catalyst information with the original dataset, selects specific 
    columns based on the reaction type, and identifies rows with the maximum target values for 
    each iteration. It also tracks changes in the maximum target values across iterations.

    Parameters
    ----------
    condition : dict
        A dictionary containing configuration details. Must include:
        - 'Reaction': str, the type of reaction (e.g., 'CH3OH_Batch').
        - 'target_name': str, the name of the target column to maximize.
    converted : dict
        A dictionary containing processed data. Must include:
        - 'data': pd.DataFrame, the main dataset containing experimental results.
    data_sort : pd.DataFrame
        A DataFrame containing sorted catalyst information with a `Catalyst` column.

    Returns
    -------
    tuple
        A tuple containing:
        - df_sort_max : pd.DataFrame
            The merged and sorted DataFrame with columns specific to the reaction type.
        - df_max_target : pd.DataFrame
            A DataFrame containing rows with the maximum target value for each iteration.
        - df_jump_cat : pd.DataFrame
            A DataFrame containing unique rows from `df_max_target`, representing jumps in the 
            maximum target value across iterations.

    Notes
    -----
    - For each reaction type, specific columns are selected from the dataset.
    - The function iterates through all iterations to find rows with the maximum target values.
    - Tracks jumps in the maximum target value across iterations and removes duplicates in the result.
    - The columns `No.` and `Iteration` in the result are converted to integers.
    """
    Reaction = condition['Reaction']
    target_name = condition['target_name']
    data = converted['data']

    # Merge the original data with the sorted 'Catalyst' column
    df_sort_max = pd.concat([data, data_sort['Catalyst']], axis=1)

    if Reaction == 'CH3OH_Batch':
        df_sort_max = df_sort_max.loc[:, [
            'No.', 'Catalyst',
            'PGM1', 'PGM1_wt%', 'PGM2', 'PGM2_wt%', 'PGM3', 'PGM3_wt%',
            'Ad1', 'Ad1_wt%', 'Ad2', 'Ad2_wt%', 'Ad3', 'Ad3_wt%', 'Ad4', 'Ad4_wt%',
            'Support_name', 'Calc T',
            target_name, 'Iteration'
        ]]


    # Initialize DataFrame to store maximum target values
    df_max_target = pd.DataFrame(columns=df_sort_max.columns, dtype='float64')
    num_loop = {}

    # Start by selecting data from iteration 0
    idx = df_sort_max['Iteration'] == 0
    num_loop[0] = df_sort_max[idx]

    # Get the entry with the maximum target value from iteration 0
    df_max_target = pd.concat([df_max_target, num_loop[0].loc[[num_loop[0][target_name].idxmax()]]], ignore_index=True)

    # Get the maximum number of iterations
    max_loop_num = df_sort_max['Iteration'].max()

    # Iterate through each subsequent iteration
    for i in range(max_loop_num):
        idx = df_sort_max['Iteration'] == i + 1
        num_loop[i + 1] = df_sort_max[idx]

        # Compare the maximum target value in the current iteration with the overall maximum
        if num_loop[i + 1][target_name].max() >= df_max_target[target_name].max():
            df_max_target = pd.concat([df_max_target, num_loop[i + 1].loc[[num_loop[i + 1][target_name].idxmax()]]], ignore_index=True)
        else:
            df_max_target = pd.concat([df_max_target, df_max_target.loc[[df_max_target[target_name].idxmax()]]], ignore_index=True)

    # Convert 'No.' and 'Iteration' columns to integers
    df_max_target.loc[:, ['No.', 'Iteration']] = df_max_target.loc[:, ['No.', 'Iteration']].astype(int)

    # Remove duplicate entries from the result
    df_jump_cat = df_max_target.drop_duplicates()

    return df_sort_max, df_max_target, df_jump_cat




def plot_prediction(condition: dict, target_true: np.array, target_pred: np.array, 
                    rmse_value: np.array, r2_value: np.array, color='blue', 
                    fig_name='prediction_plot.svg', label_name: str = y_label_name, save: bool = False):
    """
    Plot predicted vs. true target values for model evaluation.

    This function generates a scatter plot comparing predicted and true target values, 
    displaying the RMSE and R² values as annotations. The plot can optionally be saved 
    in both SVG and PDF formats.

    Parameters
    ----------
    condition : dict
        A dictionary containing configuration details. Must include:
        - 'date': str, the date for naming output files.
        - 'ML_model': str, the machine learning model name.
        - 'pgm_model': int, identifier for the PGM model.
        - 'add_model': int, identifier for the additive model.
        - 'supp_model': int, identifier for the support model.
    target_true : np.array
        Array of true target values from the experimental data.
    target_pred : np.array
        Array of predicted target values from the model.
    rmse_value : float
        Root mean squared error (RMSE) of the model predictions.
    r2_value : float
        Coefficient of determination (R²) of the model predictions.
    color : str, optional
        The color of the scatter plot points. Default is 'blue'.
    fig_name : str, optional
        Name of the output file for the plot. Default is 'prediction_plot.svg'.
    label_name : str, optional
        Label for the y-axis (e.g., target variable name). Default is `y_label_name`.
    save : bool, optional
        Whether to save the plot as SVG and PDF files. Default is False.

    Returns
    -------
    None

    Notes
    -----
    - The plot includes a diagonal reference line to indicate perfect predictions.
    - RMSE and R² values are displayed on the plot as annotations.
    - If `save` is True, the plot is saved in both SVG and PDF formats in the directories specified by `DIR_svg` and `DIR_pdf`.
    """   
    date, ML_model = condition['date'], condition['ML_model']
    pgm_model = condition['pgm_model']
    add_model = condition['add_model']
    supp_model = condition['supp_model']

    fig, ax = plt.subplots(facecolor='white', figsize=(8, 8))
    bbox_dict = dict(edgecolor='black', fill=False)
    min_num = min(min(target_true), min(target_pred))
    max_num = max(max(target_true), max(target_pred))

    ax.plot([min_num * 1.1, max_num * 1.1], [min_num * 1.1, max_num * 1.1], linestyle="--", c="gray")
    ax.plot(target_true, target_pred, "o", color=color, alpha=0.5)
    ax.text(0.05, 0.85, f'RMSE: {rmse_value:.3f}\n$R^{2}$: {r2_value:.3f}', bbox=bbox_dict, transform=ax.transAxes)

    ax.set_xlim(min_num * 1.1, max_num * 1.1)
    ax.set_ylim(min_num * 1.1, max_num * 1.1)
    ax.set_aspect('equal')
    ax.set_xlabel('Experimental' + ' ' + label_name)
    ax.set_ylabel('Predicted' + ' ' + label_name)

    if save:
        plt.savefig(f'{DIR_svg}/{date}_{Reaction}_{ML_model}_{fig_name}_prop{pgm_model}{add_model}{supp_model}.svg',
                    bbox_inches="tight", dpi=dpi_value)
        plt.savefig(f'{DIR_pdf}/{date}_{Reaction}_{ML_model}_{fig_name}_prop{pgm_model}{add_model}{supp_model}.pdf',
                    bbox_inches="tight", dpi=dpi_value)

    del fig, ax
    plt.close()




def grid_search_CV(feat: pd.DataFrame, target: pd.Series,
                   split_type: str = 'K', split_num: int = 10, 
                   use_model: str = 'ETR', plot: bool = False) -> tuple:
    """
    Perform grid search with cross-validation for model hyperparameter tuning.

    This function applies a grid search on specified models using cross-validation 
    to find the best hyperparameters for the given features and target data.

    Parameters
    ----------
    feat : pd.DataFrame
        The feature matrix (independent variables).
    target : pd.Series
        The target variable (dependent variable).
    split_type : str, optional
        The type of cross-validation splitting strategy. Default is 'K'.
        - 'K': K-Fold cross-validation.
        - 'SS-0.1': ShuffleSplit with a test size of 0.1.
        - 'SS-0.2': ShuffleSplit with a test size of 0.2.
        - 'SS-0.25': ShuffleSplit with a test size of 0.25.
    split_num : int, optional
        The number of splits for cross-validation. Default is 10.
    use_model : str, optional
        The model to use for grid search. Default is 'ETR'.
        - 'ETR': ExtraTreesRegressor.
        - 'RFR': RandomForestRegressor.
    plot : bool, optional
        Whether to generate a plot of the cross-validation results. Default is False.

    Returns
    -------
    tuple
        A tuple containing:
        - cvmodel : GridSearchCV
            The grid search object containing the best model and parameters.
        - ret : dict
            The results of the cross-validation, including metrics such as RMSE and R².

    Notes
    -----
    - For 'ETR' (ExtraTreesRegressor) and 'RFR' (RandomForestRegressor), the grid search 
      explores the following hyperparameters:
        - `n_estimators`: [100, 500, 1500]
        - `bootstrap`: [True, False]
    - The cross-validation is conducted using the specified splitting strategy (`cvf`).
    - The `crossvalid` function is used to perform cross-validation and return evaluation metrics.
    """ 
    if split_type == 'K':
        cvf = KFold(n_splits=split_num, shuffle=True, random_state=1107)
    elif split_type == 'SS-0.1':
        cvf = ShuffleSplit(n_splits=split_num, random_state=1107, test_size=0.1)
    elif split_type == 'SS-0.2':
        cvf = ShuffleSplit(n_splits=split_num, random_state=1107, test_size=0.2)
    elif split_type == 'SS-0.25':
        cvf = ShuffleSplit(n_splits=split_num, random_state=1107, test_size=0.25)

    if 'ETR' == use_model:
        cvmodel = GridSearchCV(
            ExtraTreesRegressor(n_jobs=1, random_state=1107),
            param_grid={"n_estimators": [100, 500, 1000, 1500], 'bootstrap': [True, False]},
            n_jobs=4, cv=5)
        ret = crossvalid(feat, target, cvmodel, cvf, plot=plot)

    elif 'RFR' == use_model:
        cvmodel = GridSearchCV(
            RandomForestRegressor(n_jobs=1, random_state=1107),
            param_grid={"n_estimators": [100, 500, 1000, 1500], 'bootstrap': [True, False]},
            n_jobs=4, cv=5)
        ret = crossvalid(feat, target, cvmodel, cvf, plot=plot)
        
    return cvmodel, ret




def crossvalid(feat: pd.DataFrame, target: pd.Series,
               model: object, cvf: object, plot: bool = False) -> dict:
    """
    Perform cross-validation for a given model and dataset.

    This function evaluates a model's performance using cross-validation and calculates 
    various metrics such as RMSE, MSE, MAE, and R² for both training and testing datasets.
    Optionally, it can generate plots for predicted vs. true target values.

    Parameters
    ----------
    feat : pd.DataFrame
        The feature matrix (independent variables).
    target : pd.Series
        The target variable (dependent variable).
    model : object
        The machine learning model to evaluate (e.g., a regressor).
    cvf : object
        A cross-validation splitting object (e.g., KFold, ShuffleSplit).
    plot : bool, optional
        Whether to generate plots for training and testing predictions. Default is False.

    Returns
    -------
    dict
        A dictionary containing the average cross-validation metrics:
        - 'trn_rmse': float, root mean squared error for training data.
        - 'trn_mse': float, mean squared error for training data.
        - 'trn_mae': float, mean absolute error for training data.
        - 'trn_r2': float, R² score for training data.
        - 'tes_rmse': float, root mean squared error for testing data.
        - 'tes_mse': float, mean squared error for testing data.
        - 'tes_mae': float, mean absolute error for testing data.
        - 'tes_r2': float, R² score for testing data.

    Notes
    -----
    - The function uses `cvf.split` to generate training and testing splits.
    - The metrics are calculated for each fold, and their averages are reported.
    - RMSE, MSE, and MAE are calculated using `mean_squared_error` and `mean_absolute_error`.
    - R² is calculated using `r2_score`.
    """
    mse_trn, mse_tes = [], []
    mae_trn, mae_tes = [], []
    r_2_tes, r_2_trn = [], []

    for train_index, test_index in cvf.split(feat):
        feat_trn, feat_tes = np.array(feat)[train_index], np.array(feat)[test_index]
        target_trn, target_tes = np.array(target)[train_index], np.array(target)[test_index]

        model.fit(feat_trn, target_trn)
        target_trn_pred, target_tes_pred = model.predict(feat_trn), model.predict(feat_tes)

        mse_tes.append(mean_squared_error(target_tes_pred, target_tes))
        mse_trn.append(mean_squared_error(target_trn_pred, target_trn))
        mae_tes.append(mean_absolute_error(target_tes_pred, target_tes))
        mae_trn.append(mean_absolute_error(target_trn_pred, target_trn))
        r_2_tes.append(r2_score(target_tes, target_tes_pred))
        r_2_trn.append(r2_score(target_trn, target_trn_pred))
        rmse_tes = np.sqrt(np.array(mse_tes))
        rmse_trn = np.sqrt(np.array(mse_trn))

        if plot == True:
            plot_prediction(condition, target_trn, target_trn_pred,
                            rmse_value=round(np.array(rmse_trn).mean(), 2),
                            r2_value=round(np.array(r_2_trn).mean(), 2),
                            color='blue', fig_name='train', label_name=y_label_name)
            
            plot_prediction(condition, target_tes, target_tes_pred,
                            rmse_value=round(np.array(rmse_tes).mean(), 2),
                            r2_value=round(np.array(r_2_tes).mean(), 2),
                            color='red', fig_name='test', label_name=y_label_name)

    rmse_tes = np.sqrt(np.array(mse_tes))
    rmse_trn = np.sqrt(np.array(mse_trn))

    print("Train ... RMSE: %1.3f, MSE: %1.3f, MAE: %1.3f, R2: %1.3f, RMSE detail (sd: %1.3f, min:%1.3f, max:%1.3f)"
          % (rmse_trn.mean(), np.array(mse_trn).mean(), np.array(mae_trn).mean(),
             np.array(r_2_trn).mean(), rmse_trn.std(), rmse_trn.min(), rmse_trn.max()))
    print("Test ... RMSE: %1.3f, MSE: %1.3f, MAE: %1.3f, R2: %1.3f, RMSE detail (sd: %1.3f, min:%1.3f, max:%1.3f)"
          % (rmse_tes.mean(), np.array(mse_tes).mean(), np.array(mae_tes).mean(),
             np.array(r_2_tes).mean(), rmse_tes.std(), rmse_tes.min(), rmse_tes.max()))

    ret = {}
    ret['trn_rmse'] = rmse_trn.mean()
    ret['trn_mse'] = np.array(mse_trn).mean()
    ret['trn_mae'] = np.array(mae_trn).mean()
    ret['trn_r2'] = np.array(r_2_trn).mean()
    ret['tes_rmse'] = rmse_tes.mean()
    ret['tes_mse'] = np.array(mse_tes).mean()
    ret['tes_mae'] = np.array(mae_tes).mean()
    ret['tes_r2'] = np.array(r_2_tes).mean()
    return ret




def crossvalid_err(feat: pd.DataFrame, target: pd.Series,
                   data_mol: pd.DataFrame, model: object, cvf: object) -> tuple:
    """
    Perform cross-validation with residual error analysis.

    This function evaluates a model's performance using cross-validation, calculating metrics 
    such as RMSE, MSE, MAE, and R² for both training and testing datasets. Additionally, it 
    computes and stores the residual errors for testing data.

    Parameters
    ----------
    feat : pd.DataFrame
        The feature matrix (independent variables).
    target : pd.Series
        The target variable (dependent variable).
    data_mol : pd.DataFrame
        A DataFrame containing additional metadata, including catalyst names, for residual error analysis.
    model : object
        The machine learning model to evaluate (e.g., a regressor).
    cvf : object
        A cross-validation splitting object (e.g., KFold, ShuffleSplit).

    Returns
    -------
    tuple
        A tuple containing:
        - ret : dict
            A dictionary of average cross-validation metrics:
            - 'trn_rmse': float, root mean squared error for training data.
            - 'trn_mse': float, mean squared error for training data.
            - 'trn_mae': float, mean absolute error for training data.
            - 'trn_r2': float, R² score for training data.
            - 'tes_rmse': float, root mean squared error for testing data.
            - 'tes_mse': float, mean squared error for testing data.
            - 'tes_mae': float, mean absolute error for testing data.
            - 'tes_r2': float, R² score for testing data.
        - res_err_sum : pd.DataFrame
            A DataFrame containing residual error analysis for testing data. Columns include:
            - 'test_index': Test sample indices.
            - 'Catalyst': Catalyst names.
            - 'test_target_true': True target values for testing data.
            - 'test_target_pred': Predicted target values for testing data.
            - 'abs(Residual_error)': Absolute residual error for testing data.

    Notes
    -----
    - The function uses `cvf.split` to generate training and testing splits.
    - Metrics are calculated for each fold, and their averages are reported.
    - Residual errors are calculated for testing data and stored in `res_err_sum`, 
      which is sorted by the absolute residual error in descending order.
    """
    mse_trn, mse_tes = [], []
    mae_trn, mae_tes = [], []
    r_2_tes, r_2_trn = [], []
    
    res_err_sum = pd.DataFrame()
    for train_index, test_index in cvf.split(feat):
        feat_trn, feat_tes = np.array(feat)[train_index], np.array(feat)[test_index]
        target_trn, target_tes = np.array(target)[train_index], np.array(target)[test_index]

        model.fit(feat_trn, target_trn)
        target_trn_pred, target_tes_pred = model.predict(feat_trn), model.predict(feat_tes)

        mse_tes.append(mean_squared_error(target_tes_pred, target_tes))
        mse_trn.append(mean_squared_error(target_trn_pred, target_trn))
        mae_tes.append(mean_absolute_error(target_tes_pred, target_tes))
        mae_trn.append(mean_absolute_error(target_trn_pred, target_trn))
        r_2_tes.append(r2_score(target_tes, target_tes_pred))
        r_2_trn.append(r2_score(target_trn, target_trn_pred))
        rmse_tes = np.sqrt(np.array(mse_tes))
        rmse_trn = np.sqrt(np.array(mse_trn))
        
        res_err = pd.DataFrame({
            'test_index': test_index+1,
            'Catalyst': data_mol.iloc[test_index]['Catalyst'],
            'test_target_true': target_tes,
            'test_target_pred': target_tes_pred,
            'abs(Residual_error)': abs(target_tes_pred-target_tes),
            }).reset_index(drop=True)

        res_err_sum = pd.concat([res_err_sum, res_err])
        res_err_sum.drop_duplicates(subset=['Catalyst', 'abs(Residual_error)'], keep='last', inplace=True)
        res_err_sum.sort_values('abs(Residual_error)', ascending=False, inplace=True)

    rmse_tes = np.sqrt(np.array(mse_tes))
    rmse_trn = np.sqrt(np.array(mse_trn))

    print("Train ... RMSE: %1.3f, MSE: %1.3f, MAE: %1.3f, R2: %1.3f, RMSE detail (sd: %1.3f, min:%1.3f, max:%1.3f)"
          % (rmse_trn.mean(), np.array(mse_trn).mean(), np.array(mae_trn).mean(), np.array(r_2_trn).mean(), 
             rmse_trn.std(), rmse_trn.min(), rmse_trn.max()))
    print("Test ... RMSE: %1.3f, MSE: %1.3f, MAE: %1.3f, R2: %1.3f, RMSE detail (sd: %1.3f, min:%1.3f, max:%1.3f)"
          % (rmse_tes.mean(), np.array(mse_tes).mean(), np.array(mae_tes).mean(), np.array(r_2_tes).mean(),
             rmse_tes.std(), rmse_tes.min(), rmse_tes.max()))

    ret = {}
    ret['trn_rmse'] = rmse_trn.mean()
    ret['trn_mse'] = np.array(mse_trn).mean()
    ret['trn_mae'] = np.array(mae_trn).mean()
    ret['trn_r2'] = np.array(r_2_trn).mean()
    ret['tes_rmse'] = rmse_tes.mean()
    ret['tes_mse'] = np.array(mse_tes).mean()
    ret['tes_mae'] = np.array(mae_tes).mean()
    ret['tes_r2'] = np.array(r_2_tes).mean()
    return ret, res_err_sum




def one_shot_plot(model:object, feat:pd.DataFrame, target:pd.Series, test_size:float=0.2,
                  random_state:int=1107, fig_size:tuple=(8, 8), fontsize:int=20,
                  train_color:str='blue', test_color:str='red',
                  xy_min:int=0, xy_max:int=40, tick_interval:int=10,
                  alpha_value:float=0.7, dot_size:int=7,
                  save:bool=False, fig_name:str='one_shot_plot'):
    """
    Create a scatter plot of predicted vs. experimental values for a one-shot train-test split.

    This function trains a given model on a randomly split training dataset and tests it on 
    the test dataset. It visualizes the experimental vs. predicted values for both the training 
    and test datasets on the same plot.

    Parameters
    ----------
    model : object
        The machine learning model to train and evaluate.
    feat : pd.DataFrame
        The feature matrix (independent variables).
    target : pd.Series
        The target variable (dependent variable).
    test_size : float, optional
        The proportion of the dataset to include in the test split. Default is 0.2.
    random_state : int, optional
        Random seed for reproducibility of the train-test split. Default is 1107.
    fig_size : tuple, optional
        The size of the figure. Default is (8, 8).
    fontsize : int, optional
        Font size for labels, legend, and ticks. Default is 20.
    train_color : str, optional
        Color for training data points. Default is 'blue'.
    test_color : str, optional
        Color for testing data points. Default is 'red'.
    xy_min : int, optional
        Minimum value for both x and y axes. Default is 0.
    xy_max : int, optional
        Maximum value for both x and y axes. Default is 40.
    tick_interval : int, optional
        Interval between ticks on the x and y axes. Default is 10.
    alpha_value : float, optional
        Transparency of the points. Default is 0.7.
    dot_size : int, optional
        Size of the scatter plot markers. Default is 7.
    save : bool, optional
        Whether to save the plot as SVG and PDF files. Default is False.
    fig_name : str, optional
        The name of the output file (if saved). Default is 'one_shot_plot'.

    Returns
    -------
    None

    Notes
    -----
    - The plot includes data points for both training and testing datasets, as well as a diagonal 
      reference line indicating perfect predictions.
    - If `save` is True, the plot is saved in both SVG and PDF formats to the specified directories.
    """
    plt.figure(facecolor='white', figsize=fig_size)
    plt.subplot().set_aspect('equal')

    x_train, x_test, y_train, y_test \
        = train_test_split(feat, target, test_size=test_size, random_state=random_state)
    
    model.fit(x_train, y_train)
    y_train_pred, y_test_pred = model.predict(x_train), model.predict(x_test)

    plt.plot(y_test, y_test_pred, 'o', color=test_color,
             markersize=dot_size, alpha=alpha_value, zorder=2, label='Test')
    plt.plot(y_train, y_train_pred, 'o', color=train_color,
             markersize=dot_size-1, alpha=alpha_value-0.1, zorder=1, label='Training')

    plt.plot([-0.2, int(target.max())*1.5], [-0.2, int(target.max())*1.5],
             color='black', alpha=0.8, zorder=0, linestyle='-', linewidth=1.0)
    plt.xlim([xy_min, xy_max])
    plt.ylim([xy_min, xy_max])
    plt.legend(title=f'Test size = {test_size}',
               title_fontsize=fontsize+1, fontsize=fontsize,
               frameon=True, markerscale=1.1, handletextpad=0.5)
    plt.xticks(np.arange(xy_min, xy_max + tick_interval, tick_interval),
               fontsize=fontsize-1)
    plt.yticks(np.arange(xy_min, xy_max + tick_interval, tick_interval),
               fontsize=fontsize-1)

    plt.xlabel('Experimental ' + y_label_name, fontsize=fontsize)
    plt.ylabel('Predicted ' + y_label_name, fontsize=fontsize)

    if save is not False:
        plt.savefig(f'{DIR_svg}/{fig_name}.svg', bbox_inches="tight", dpi=dpi_value)
        plt.savefig(f'{DIR_pdf}/{fig_name}.pdf', bbox_inches="tight", dpi=dpi_value)




def plot_importance(model:object, labels:list, topk:int=20,
                    tick_interval:int=0.1,
                    fig_size:tuple=(8, 8), fontsize:int=18,
                    color:str='blue', alpha_value:float=0.6,
                    save:bool=False, fig_name:str='Feature_importance'):
    """
    Plot feature importance for a trained model.

    This function generates a horizontal bar plot to visualize the importance of the top `k` 
    features based on a model's `feature_importances_` attribute.

    Parameters
    ----------
    model : object
        A trained machine learning model with a `feature_importances_` attribute (e.g., tree-based models).
    labels : list
        A list of feature names corresponding to the columns in the dataset used for training.
    topk : int, optional
        The number of top features to display in the plot. Default is 20.
    tick_interval : float, optional
        The interval for x-axis ticks representing the feature importance scores. Default is 0.1.
    fig_size : tuple, optional
        The size of the figure. Default is (8, 8).
    fontsize : int, optional
        Font size for labels, ticks, and titles. Default is 18.
    color : str, optional
        Color of the bars in the plot. Default is 'blue'.
    alpha_value : float, optional
        Transparency of the bars. Default is 0.6.
    save : bool, optional
        Whether to save the plot as SVG and PDF files. Default is False.
    fig_name : str, optional
        The name of the output file (if saved). Default is 'Feature_importance'.

    Returns
    -------
    None

    Notes
    -----
    - The `feature_importances_` attribute of the model must be available and populated (e.g., after calling `model.fit`).
    - The plot shows the top `k` features, sorted by their importance scores.
    - If `save` is True, the plot is saved in both SVG and PDF formats to the specified directories.
    """
    plt.figure(facecolor='white', figsize=fig_size)

    # feature_importances_: モデルが訓練 (model.fit(feat, target)) された後に自動的に生成される属性で、
    # 各特徴量が予測にどれだけ影響を与えたかを数値的に示す。
    importances = model.feature_importances_
    indices = np.argsort(importances)
    topk_idx = indices[-topk:]
    plt.barh(range(len(topk_idx)),
             importances[topk_idx], color=color, alpha=alpha_value, align='center')
    
    plt.xticks(np.arange(0, round(importances[topk_idx].max(), 1)*1.1, tick_interval), fontsize=fontsize-1)
    plt.yticks(range(len(topk_idx)), labels[topk_idx], fontsize=fontsize-1)

    plt.ylim([-1, len(topk_idx)])
    plt.xlabel("Feature importance score", fontsize=fontsize)

    if save:
        plt.savefig(f'{DIR_svg}/{fig_name}.svg', bbox_inches="tight", dpi=dpi_value)
        plt.savefig(f'{DIR_pdf}/{fig_name}.pdf', bbox_inches="tight", dpi=dpi_value)




def shap_summary_plot(model:object, feat:pd.DataFrame, target:pd.Series,
                      num_desc:int=15, save:bool=False, fig_name:str='SHAP_sumplot'):
    """
    Generate a SHAP summary plot to visualize feature importance and contributions.

    This function calculates SHAP values for the given model and dataset and generates a SHAP 
    summary plot, which shows the feature importance and the distribution of SHAP values 
    for the top features.

    Parameters
    ----------
    model : object
        A machine learning model compatible with SHAP (e.g., tree-based models like RandomForest, XGBoost).
    feat : pd.DataFrame
        The feature matrix (independent variables).
    target : pd.Series
        The target variable (dependent variable).
    num_desc : int, optional
        The number of top features to display in the SHAP summary plot. Default is 15.
    save : bool, optional
        Whether to save the plot as an SVG file. Default is False.
    fig_name : str, optional
        The name of the output file (if saved). Default is 'SHAP_sumplot'.

    Returns
    -------
    None

    Notes
    -----
    - The SHAP summary plot provides a global interpretation of the model by showing:
        - The feature importance (based on SHAP values).
        - The distribution of SHAP values for each feature (color-coded by feature value).
    - The plot is generated using `shap.summary_plot` with the following configurations:
        - The top `num_desc` features are displayed.
        - The plot uses a dot plot style with automatic plot size adjustment.
    - If `save` is True, the plot is saved as an SVG file.
    """
    model.fit(feat, target)  # Fit the model to the features and target
    explainer = shap.TreeExplainer(model=model)  # Create a SHAP explainer object for tree-based models
    shap_values = explainer.shap_values(feat)  # Compute SHAP values for the feature set

    # Generate the SHAP summary plot
    shap.summary_plot(
        shap_values,  features=feat, feature_names=feat.columns, max_display=num_desc,
        plot_type='dot', color=None, axis_color='k', title=None, alpha=0.8,
        show=False, sort=True, color_bar=True, plot_size='auto', layered_violin_max_num_bins=20,
        class_names=None, class_inds=None, color_bar_label='Feature value',
        # cmap='cool',
        # cmap='cool_r',
        auto_size_plot=None, use_log_scale=False)

    if save:
        input_svg = f'{DIR_svg}/{fig_name}.svg'
        input_png = f'{DIR_svg}/{fig_name}.png'
        plt.savefig(input_svg, bbox_inches='tight', dpi=dpi_value)
        plt.savefig(input_png, bbox_inches='tight', dpi=dpi_value)

        output_pdf = f'{DIR_pdf}/{fig_name}.pdf'
        img = Image.open(input_png)
        cov_pdf = img2pdf.convert(input_png)
        pdf_file = open(output_pdf , "wb")
        pdf_file.write(cov_pdf)
        img.close()
        pdf_file.close()



def loo_plot(model:object, feat:pd.DataFrame, target:pd.Series,
             feat_cols:list, out_range:float=0.0,
             xy_min:int=0, xy_max:int=40, tick_interval:int=10,
             fig_size:tuple=(8,8), fontsize:int=20, text_fontsize:int=7,
             color:str='red', alpha_value:float=0.7, dot_size:int=25,
             save:bool=False, fig_name:str='LeaveOneOut_plot') -> pd.DataFrame:
    """
    Generate a scatter plot for Leave-One-Out Cross-Validation (LOOCV) predictions.

    This function performs LOOCV on the given dataset using the specified model, calculates 
    prediction errors, and creates a scatter plot comparing experimental and predicted values. 
    Points outside a specified error range are annotated with their indices.

    Parameters
    ----------
    model : object
        A machine learning model compatible with the `fit` and `predict` methods.
    feat : pd.DataFrame
        The feature matrix (independent variables).
    target : pd.Series
        The target variable (dependent variable).
    feat_cols : list
        A list of column names to be used as features for training the model.
    out_range : float, optional
        The allowable error range for annotating outlier points. Default is 0.0, which sets the range 
        to 10% of `xy_max`.
    xy_min : int, optional
        Minimum value for both x and y axes. Default is 0.
    xy_max : int, optional
        Maximum value for both x and y axes. Default is 40.
    tick_interval : int, optional
        Interval between ticks on the x and y axes. Default is 10.
    fig_size : tuple, optional
        The size of the figure. Default is (8, 8).
    fontsize : int, optional
        Font size for axis labels and ticks. Default is 20.
    text_fontsize : int, optional
        Font size for annotations. Default is 7.
    color : str, optional
        Color for scatter plot points. Default is 'red'.
    alpha_value : float, optional
        Transparency of the scatter plot points. Default is 0.7.
    dot_size : int, optional
        Size of the scatter plot markers. Default is 25.
    save : bool, optional
        Whether to save the plot as SVG and PDF files. Default is False.
    fig_name : str, optional
        The name of the output file (if saved). Default is 'LeaveOneOut_plot'.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the following columns:
        - 'original_target': The true target values.
        - 'pred_target': The predicted target values.
        - 'Error': Absolute error between predictions and true values.
        - 'Squared Error': Squared error between predictions and true values.
        - 'RMSE': Root mean squared error for each prediction.

    Notes
    -----
    - The plot includes a reference line (y = x) and dashed lines for the error range.
    - Points outside the specified error range are annotated with their indices.
    - If `save` is True, the plot is saved in both SVG and PDF formats to the specified directories.
    """
    loo = LeaveOneOut()

    feat_loo = feat.copy()
    for train_index, test_index in loo.split(feat):
        X_train, X_test = feat.loc[train_index, feat_cols], feat.loc[test_index, feat_cols]
        y_train, y_test = target.iloc[train_index], target.iloc[test_index]
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)
        feat_loo.loc[test_index, 'original_target'] = y_test
        feat_loo.loc[test_index, 'pred_target'] = y_pred
        feat_loo.loc[test_index, 'Error'] = abs(y_pred-y_test)
        feat_loo.loc[test_index, 'Squared Error'] = abs(y_pred-y_test)**2
        feat_loo.loc[test_index, 'RMSE'] = math.sqrt(mean_squared_error(y_test, y_pred))

    plt.figure(facecolor='white', figsize=fig_size)

    if out_range == 0.0:
        out_range = abs(xy_max)*0.1
    else:
        out_range = out_range

    x = np.arange(-xy_max*2, xy_max*2, 0.1)
    y = x
    plt.plot(x, y,  linewidth=1, color='black', alpha=0.8, zorder=0)
    plt.plot(x, y+out_range,  linewidth=0.8, linestyle='dashed',
             color='black', alpha=0.8, zorder=0)
    plt.plot(x, y-out_range,  linewidth=0.8, linestyle='dashed',
             color='black', alpha=0.8, zorder=0)

    plt.scatter(feat_loo.loc[:, 'original_target'],
                feat_loo.loc[:, 'pred_target'], color=color, s=dot_size,
                alpha=alpha_value, zorder=1)
    
    plt.xlim([xy_min, xy_max])
    plt.ylim([xy_min, xy_max])
    
    plt.xticks(np.arange(xy_min, xy_max + tick_interval, tick_interval),
               fontsize=fontsize-1)
    plt.yticks(np.arange(xy_min, xy_max + tick_interval, tick_interval),
               fontsize=fontsize-1)
    
    plt.xlabel('Experimental ' + y_label_name, fontsize=fontsize)
    plt.ylabel('Predicted ' + y_label_name, fontsize=fontsize)

    texts = [
        plt.text(feat_loo.loc[i, 'original_target'],
                 feat_loo.loc[i, 'pred_target'], f"{i+1}", fontsize=text_fontsize, zorder=2)
        for i in range(len(feat))
        if feat_loo.loc[i, 'pred_target'] <= feat_loo.loc[i, 'original_target'] - out_range or
        feat_loo.loc[i, 'pred_target'] >= feat_loo.loc[i, 'original_target'] + out_range
    ]
    adjust_text(texts, arrowprops=dict(arrowstyle='-', color='black'))

    if save:
        plt.savefig(f'{DIR_svg}/{fig_name}.svg', bbox_inches="tight", dpi=dpi_value)
        plt.savefig(f'{DIR_pdf}/{fig_name}.pdf', bbox_inches="tight", dpi=dpi_value)
    return feat_loo



def cv_check(condition:dict, data_cols:list, desc_cols:list,
             model:object, shuffle_mode:str='10-fold') -> tuple:
    """
    Evaluate combinations of model configurations using cross-validation.

    This function performs cross-validation on different combinations of models for 
    Platinum Group Metals (PGM), additives, and supports, and 
    evaluates their performance metrics (RMSE, MAE, R²).

    Parameters
    ----------
    condition : dict
        A dictionary containing configuration details. Must include:
        - 'pgm_num': int, number of PGM elements.
        - 'add_num': int, number of additive elements.
        - 'supp_num': int, number of support elements.
    data_cols : list
        A list specifying the columns in the dataset relevant to features and weights.
    desc_cols : list
        A list specifying descriptor columns for the dataset.
    model : object
        A machine learning model compatible with `fit` and `predict` methods (e.g., tree-based models).
    shuffle_mode : str, optional
        The cross-validation strategy to use. Default is '10-fold'.
        - 'ShuffleSplit_01': ShuffleSplit with 10% test size.
        - 'ShuffleSplit_02': ShuffleSplit with 20% test size.
        - '5-fold': 5-fold cross-validation.
        - '10-fold': 10-fold cross-validation.

    Returns
    -------
    tuple
        A tuple containing:
        - export : pd.DataFrame
            A DataFrame with the results for each combination of model configurations:
            - 'prop': str, the combination identifier (e.g., '12' for add_model=1, supp_model=2).
            - 'RMSE': float, root mean squared error.
            - 'MAE': float, mean absolute error.
            - 'R2': float, R² score.
        - prop_list : list
            A list of all tested combinations of model configurations.

    Notes
    -----
    - For PGM, PRM, additives, and supports, the function evaluates all combinations 
      of their models (0, 1, 2).
    - The cross-validation strategy is determined by `shuffle_mode`.
    - The function uses the `crossvalid` function to calculate evaluation metrics.
    """

    prop_list = list(itertools.product([0, 1, 2], repeat=3))  # Generate all combinations
    columns = ['prop', 'RMSE', 'MAE', 'R2']
    indexs = np.arange(len(prop_list))  # Create index based on the number of combinations
    export = pd.DataFrame(columns=columns, index=indexs)  # Create DataFrame to store results

    for i in range(len(prop_list)):
        print(f'prop: {prop_list[i]}')
        pgm_model = prop_list[i][0]
        add_model = prop_list[i][1]
        supp_model = prop_list[i][2]

        # Convert the data based on current model combination
        converted = converter.data_convert(condition, data_cols, desc_cols, sheet_name='data_sheet_name',
                                            select_desc=True,
                                            use_models=[pgm_model, add_model, supp_model],
                                            local=False, idx=None)
        feat, target = converted['feat'], converted['target']

        # Choose cross-validation strategy
        if shuffle_mode == 'ShuffleSplit_01':
            cvf = ShuffleSplit(n_splits=100, random_state=1107, test_size=0.1)
        elif shuffle_mode == 'ShuffleSplit_02':
            cvf = ShuffleSplit(n_splits=100, random_state=1107, test_size=0.2)
        elif shuffle_mode == '5-fold':
            cvf = KFold(n_splits=5, shuffle=True, random_state=1107)
        elif shuffle_mode == '10-fold':  # default
            cvf = KFold(n_splits=10, shuffle=True, random_state=1107)

        # Perform cross-validation and store the results
        ret = crossvalid(feat, target, model, cvf, plot=False)
        prop = f'{pgm_model}{add_model}{supp_model}'
        
        export.loc[i, 'prop'] = prop
        export.loc[i, 'R2'] = ret['tes_r2']
        export.loc[i, 'RMSE'] = ret['tes_rmse']
        export.loc[i, 'MAE'] = ret['tes_mae']

    return export, prop_list



# For each iteration, perform cross-validation using the input representations of the best model
# and the model with the same number of digits, save the results as a CSV file, and return all results in dictionary format.
def loop_CV_short(condition:dict, converted:dict, data_cols:dict, desc_cols:dict,
                  shuffle_mode:str='10-fold',
                  best_prop_list:list=[]) -> dict:
    """
    Evaluate combinations of model configurations using cross-validation.

    This function performs cross-validation on different combinations of models for 
    Platinum Group Metals (PGM), additives, and supports, and 
    evaluates their performance metrics (RMSE, MAE, R²).

    Parameters
    ----------
    condition : dict
        A dictionary containing configuration details. Must include:
        - 'pgm_num': int, number of PGM elements.
        - 'add_num': int, number of additive elements.
        - 'supp_num': int, number of support elements.
    data_cols : list
        A list specifying the columns in the dataset relevant to features and weights.
    desc_cols : list
        A list specifying descriptor columns for the dataset.
    model : object
        A machine learning model compatible with `fit` and `predict` methods (e.g., tree-based models).
    shuffle_mode : str, optional
        The cross-validation strategy to use. Default is '10-fold'.
        - 'ShuffleSplit_01': ShuffleSplit with 10% test size.
        - 'ShuffleSplit_02': ShuffleSplit with 20% test size.
        - '5-fold': 5-fold cross-validation.
        - '10-fold': 10-fold cross-validation.

    Returns
    -------
    tuple
        A tuple containing:
        - export : pd.DataFrame
            A DataFrame with the results for each combination of model configurations:
            - 'prop': str, the combination identifier (e.g., '12' for add_model=1, supp_model=2).
            - 'RMSE': float, root mean squared error.
            - 'MAE': float, mean absolute error.
            - 'R2': float, R² score.
        - prop_list : list
            A list of all tested combinations of model configurations.

    Notes
    -----
    - For PGM, Promoter, additives, and supports, the function evaluates all combinations 
      of their models (0, 1, 2).
    - The cross-validation strategy is determined by `shuffle_mode`.
    - The function uses the `crossvalid` function to calculate evaluation metrics.
    """
    CV_data_dict = {}

    date, Reaction = condition['date'], condition['Reaction']
    data = converted['data']
    num_loop_max = data['Iteration'].max()  # Maximum number of iterations

    # Define property combinations to iterate through
    prop_list = [[0, 0, 0], [1, 1, 1], [2, 2, 2], best_prop_list]

    for i in range(len(prop_list)):
        CV_data_list = []  # Store cross-validation results

        for j in range(num_loop_max + 1):  # Iterate through loops
            if len(prop_list[i]) != 0:
                # Convert data using current property list
                converted = converter.data_convert(condition, data_cols, desc_cols,
                                                   sheet_name='data_sheet_name', select_desc=True,
                                                   use_models=[prop_list[i][0], prop_list[i][1], prop_list[i][2]],
                                                   local=False, idx=j)
                data, feat, target = converted['data'], converted['feat'], converted['target']
                
                # Define and fit the model
                model = ExtraTreesRegressor(n_estimators=100, random_state=1107, n_jobs=-1)
                model.fit(feat, target)

                # Set cross-validation strategy
                if shuffle_mode == 'ShuffleSplit_01':
                    cvf = ShuffleSplit(n_splits=100, random_state=1107, test_size=0.1)
                elif shuffle_mode == 'ShuffleSplit_02':
                    cvf = ShuffleSplit(n_splits=100, random_state=1107, test_size=0.2)
                elif shuffle_mode == '5-fold':
                    cvf = KFold(n_splits=5, shuffle=True, random_state=1107)
                elif shuffle_mode == '10-fold':
                    cvf = KFold(n_splits=10, shuffle=True, random_state=1107)

                print(f'prop={prop_list[i][0]}{prop_list[i][1]}{prop_list[i][2]}', 'data_length=', len(data))
                
                # Perform cross-validation
                output = crossvalid(feat, target, model, cvf)

                # If the output is a dictionary, convert it to a DataFrame
                if isinstance(output, dict):
                    output = pd.DataFrame([output])
                CV_data_list.append(output)  # Append output to list

        # Concatenate all CV results for current property combination
        CV_data = pd.concat(CV_data_list, ignore_index=True)

        prop_key = f'prop_{prop_list[i][0]}{prop_list[i][1]}{prop_list[i][2]}'
        CV_data_dict[prop_key] = CV_data

        # Save results to CSV file
        CV_data.to_csv(f'{DIR_csv}/{date}_{Reaction}_CV_prop{prop_list[i][0]}{prop_list[i][1]}{prop_list[i][2]}_loop{num_loop_max}.csv')
    
    return CV_data_dict



# For each iteration, perform cross-validation on all combinations of input expressions, 
# save the results as a CSV file, and return all results in dictionary format.
def loop_CV_full(condition: dict, converted: dict, data_cols: dict, desc_cols: dict,
                 shuffle_mode: str = '10-fold') -> dict:
    """
    Perform exhaustive cross-validation for all combinations of property configurations.

    This function evaluates all combinations of model properties (e.g., PGM, additives, support) 
    using cross-validation for multiple iterations and saves the results for each configuration.

    Parameters
    ----------
    condition : dict
        A dictionary containing configuration details. Must include:
        - 'date': str, the date for naming output files.
        - 'Reaction': str, the reaction name for naming output files.
    converted : dict
        A dictionary containing preprocessed data. Must include:
        - 'data': pd.DataFrame, dataset with experimental information.
    data_cols : dict
        A dictionary specifying the feature and weight column names in the dataset.
    desc_cols : dict
        A dictionary specifying descriptor columns for the dataset.
    shuffle_mode : str, optional
        The cross-validation strategy to use. Default is '10-fold'.
        - 'ShuffleSplit_01': ShuffleSplit with 10% test size.
        - 'ShuffleSplit_02': ShuffleSplit with 20% test size.
        - '5-fold': 5-fold cross-validation.
        - '10-fold': 10-fold cross-validation.

    Returns
    -------
    dict
        A dictionary where keys represent property combinations (e.g., 'prop_0123') and values 
        are DataFrames containing cross-validation results for each combination.

    Notes
    -----
    - The function evaluates all combinations of model configurations (0, 1, 2) for four components:
      PGMs, Promoters, additives, and supports.
    - Cross-validation results for each property combination are saved as separate CSV files in `DIR_csv`.
    - Uses `converter.data_convert` to preprocess data for each iteration.
    - The results are stored in a dictionary with each key corresponding to a property combination.
    """
    method_list = [0, 1, 2]
    
    date, Reaction = condition['date'], condition['Reaction']
    data = converted['data']
    num_loop_max = data['Iteration'].max()

    prop_list = list(itertools.product(method_list, repeat=3))
    CV_data_dict = {}

    for i in range(len(prop_list)):
        CV_data_list = []
        for j in range(num_loop_max + 1):
            if len(prop_list[i]) != 0:
                converted = converter.data_convert(condition, data_cols, desc_cols,
                                                sheet_name='data_sheet_name', select_desc=True,
                                                use_models=[prop_list[i][0], prop_list[i][1], prop_list[i][2]],
                                                local=False, idx=j)
                data, feat, target = converted['data'], converted['feat'], converted['target']
                model = ExtraTreesRegressor(n_estimators=100, random_state=1107, n_jobs=-1)
                model.fit(feat, target)

                if shuffle_mode == 'ShuffleSplit_01':
                    cvf = ShuffleSplit(n_splits=100, random_state=1107, test_size=0.1)
                elif shuffle_mode == 'ShuffleSplit_02':
                    cvf = ShuffleSplit(n_splits=100, random_state=1107, test_size=0.2)
                elif shuffle_mode == '5-fold':
                    cvf = KFold(n_splits=5, shuffle=True, random_state=1107)
                elif shuffle_mode == '10-fold':
                    cvf = KFold(n_splits=10, shuffle=True, random_state=1107)

                print(f'prop={prop_list[i][0]}{prop_list[i][1]}{prop_list[i][2]}', 'data_length=', len(data))
                output = crossvalid(feat, target, model, cvf)
                
                # If the output is a dictionary type, convert it to a DataFrame.
                if isinstance(output, dict):
                    output = pd.DataFrame([output])  # Convert a dictionary into a single row of a DataFrame
                CV_data_list.append(output)

        CV_data = pd.concat(CV_data_list, ignore_index=True)

        prop_key = f'prop_{prop_list[i][0]}{prop_list[i][1]}{prop_list[i][2]}'
        CV_data_dict[prop_key] = CV_data

        CV_data.to_csv(f'{DIR_csv}/{date}_{Reaction}_CV_prop{prop_list[i][0]}{prop_list[i][1]}{prop_list[i][2]}_loop{num_loop_max}.csv')

    return CV_data_dict



def various_ML_CV(
    feat:pd.DataFrame, target:pd.Series,
    condition:dict, cvf_method:str='10-fold',
    log_param1:list = [1e-2, 1e-1, 1.0, 10, 1e2],
    log_param2:list = [1.0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5],
    log_param3:list = [1.0, 10, 1e2, 1e3, 1e4, 1e5]
    ) -> dict:
    """
    Perform cross-validation with various machine learning models.

    Parameters
    ----------
    feat : pd.DataFrame
        Features for training the machine learning models.
    target : pd.Series
        Target values for training the machine learning models.
    condition : dict
        Dictionary containing experiment parameters such as 'pgm_model', 'add_model', and 'supp_model'.
    cvf_method : str, optional
        Cross-validation method. Defaults to '10-fold'.
        Options: 'ShuffleSplit02', 'ShuffleSplit01', '10-fold', '5-fold'.
    log_param1 : list, optional
        Logarithmic parameter grid for regularization parameter (alpha) or epsilon for models like Lasso and SVR.
    log_param2 : list, optional
        Logarithmic parameter grid for other models' parameters like gamma or alpha for GaussianProcessRegressor.
    log_param3 : list, optional
        Logarithmic parameter grid for regularization parameter (C) for models like SVR.

    Returns
    -------
    errors : dict
        A dictionary where the keys are model names, and the values are cross-validation results.

    Notes
    -----
    - The function performs grid search with cross-validation for several models such as Lasso, Ridge, GaussianProcessRegressor (GPR),
      Support Vector Regressor (SVR), Random Forest Regressor (RFR), and Extra Trees Regressor (ETR).
    - XGBoost (XGB) and LightGBM (LGBM) are commented out but can be included for extended evaluation.
    - Cross-validation is performed using the method specified by `cvf_method`.
    """
    errors = {}
    pgm_model = condition['pgm_model']
    add_model = condition['add_model']
    supp_model = condition['supp_model']
    
    # Cross-validation method selection
    if cvf_method == 'ShuffleSplit02':
        cvf = ShuffleSplit(n_splits=100, random_state=1107, test_size=0.2)
    elif cvf_method == 'ShuffleSplit01':
        cvf = ShuffleSplit(n_splits=100, random_state=1107, test_size=0.1)
    elif cvf_method == '10-fold':
        cvf = KFold(n_splits=10, shuffle=True, random_state=1107)
    elif cvf_method == '5-fold':
        cvf = KFold(n_splits=5, shuffle=True, random_state=1107)
    
    # Lasso Regression
    print('Lasso', f'prop{pgm_model}{add_model}{supp_model}')
    cvmodel = GridSearchCV(Lasso(), param_grid={'alpha': log_param1}, n_jobs=-1)
    errors['Lasso'] = crossvalid(feat, target, cvmodel, cvf)
    
    # Ridge Regression
    print('Ridge', f'prop{pgm_model}{add_model}{supp_model}')
    cvmodel = GridSearchCV(Ridge(), param_grid={'alpha': log_param1}, n_jobs=-1)
    errors['Ridge'] = crossvalid(feat, target, cvmodel, cvf)
    
    # Gaussian Process Regressor (GPR)
    print('GPR', f'prop{pgm_model}{add_model}{supp_model}')
    kernel = gp.kernels.Matern(nu=2.5)
    cvmodel = GridSearchCV(
        gp.GaussianProcessRegressor(kernel=kernel, normalize_y=True, 
                                    n_restarts_optimizer=10, random_state=1107),
        param_grid={'alpha': log_param2}, n_jobs=-1
    )
    errors['GPR'] = crossvalid(feat, target, cvmodel, cvf)
    
    # Support Vector Regressor (SVR)
    print('SVR', f'prop{pgm_model}{add_model}{supp_model}')
    cvmodel = GridSearchCV(SVR(kernel='rbf'),
                           param_grid={'C': log_param3, 'gamma': log_param2, 'epsilon': log_param1}, n_jobs=-1)
    errors['SVR'] = crossvalid(feat, target, cvmodel, cvf)
    
    # Random Forest Regressor (RFR)
    print('RFR', f'prop{pgm_model}{add_model}{supp_model}')
    cvmodel = GridSearchCV(
        RandomForestRegressor(n_jobs=1, random_state=1107),
        param_grid={"n_estimators": [100, 500, 1000, 1500], 'bootstrap': [True, False]},
        n_jobs=4, cv=5)
    errors['RFR'] = crossvalid(feat, target, cvmodel, cvf)

    # Extra Trees Regressor (ETR)
    print('ETR', f'prop{pgm_model}{add_model}{supp_model}')
    cvmodel = GridSearchCV(
        ExtraTreesRegressor(n_jobs=1, random_state=1107),
        param_grid={"n_estimators": [100, 500, 1000, 1500], 'bootstrap': [True, False]},
        n_jobs=4, cv=5)
    errors['ETR'] = crossvalid(feat, target, cvmodel, cvf)
    
    """
    # XGBoost Regressor (XGB)
    print('XGB', f'prop{pgm_model}{prm_model}{add_model}{supp_model}')
    cvmodel = GridSearchCV(
        XGBRegressor(n_jobs=1, random_state=1107),
        param_grid={"n_estimators": [100, 250, 500, 1000, 1500]}, n_jobs=4, cv=5
    )
    errors['XGB'] = crossvalid(feat, target, cvmodel, cvf)
    
    # LightGBM Regressor (LGBM)
    print('LGBM', f'prop{pgm_model}{prm_model}{add_model}{supp_model}')
    cvmodel = GridSearchCV(
        LGBMRegressor(n_jobs=1, random_state=1107),
        param_grid={"n_estimators": [100, 250, 500, 1000, 1500]}, n_jobs=4, cv=5
    )
    errors['LGBM'] = crossvalid(feat, target, cvmodel, cvf)
    """
    
    return errors
