#!/usr/bin/env python
import sys
sys.dont_write_bytecode = True

import collections
import numpy as np
import pandas as pd
import itertools
import scipy.sparse as sparse
from scipy.stats import norm


#import modules.conditions as conditions
import modules.conditions_CH3OH_Batch as conditions

import modules.setting as setting


# 多重リストを平滑化する関数。
def flatten(l: list):
    """
    Flatten a nested list or iterable.

    Parameters
    ----------
    l : list
        A nested list or iterable.

    Yields
    ------
    el : Any
        Flattened elements of the input.
    """
    for el in l:
        if isinstance(el, collections.abc.Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el



# 任意の整数の約数一覧(昇順)を作成する。
def divisors_list(num: int) -> list:
    """
    Return a list of divisors of a given integer in ascending order.

    Parameters
    ----------
    num : int
        The integer for which to find divisors.

    Returns
    -------
    list of int
        A list of divisors of the input integer in ascending order.
    """
    divisors = []
    for i in range(1, num+1):
        if num % i == 0:
            divisors.append(i)
    return divisors



# ある整数(total_num)を、ある大きさ(size)以上で割り切れる最小の整数を計算する。
def num_jobs(total_num: int, divisors: list, size: int = int(1E6)) -> tuple:
    """
    Calculate the minimum integer that can divide `total_num` with a size greater than or equal to `size`.

    Parameters
    ----------
    total_num : int
        The total number to be divided.
    divisors : list of int
        A list of divisors to use for dividing `total_num`.
    size : int, optional
        The minimum size threshold for dividing `total_num`. Defaults to 1,000,000.

    Returns
    -------
    tuple
        A tuple containing:
        - slice_size (int): The size of each slice that divides `total_num`.
        - num_slice (int): The number of slices required to cover `total_num`.

    Notes
    -----
    - If the divided size exceeds `size`, it adjusts the slice size and number of slices accordingly.
    - If `total_num` is smaller than `size`, it returns `total_num` as the slice size with 1 as the number of slices.
    """
    catal_comb_size = []
    for i in range(len(divisors)):
        catal_comb_size.append(total_num / divisors[i])
        if catal_comb_size[i] > size:
            slice_size = int(catal_comb_size[i])
            num_slice = int(total_num / slice_size)
        if total_num < size:
            slice_size = total_num
            num_slice = 1
    return slice_size, num_slice




# 計算に用いる元素リストを作成する。
def elem_list(condition: dict, converted: dict, desc_cols: list) -> dict:

    pgm_num = condition['pgm_num']
    add_num = condition['add_num']
    supp_num = condition['supp_num']

    drop_pgms, drop_adds = condition['drop_pgms'], condition['drop_adds']
    essential_supp, drop_supp_list = condition['essential_supp'], condition['drop_supp_list']
    CalT_list = condition['CalT_list']

    pgm_desc = converted['pgm_desc']
    add_desc = converted['add_desc']
    supp_desc = converted['supp_desc']
    supp_name_dict = desc_cols['supp_name_dict']

    if pgm_num != 0:
        all_pgm_list = pgm_desc.index[1:]
        if len(drop_pgms) == 0:
            use_pgm_list = all_pgm_list
        else:
            use_pgm_list = list(set(all_pgm_list) - set(drop_pgms))
    else:
        all_pgm_list, use_pgm_list = [], []

    if add_num != 0:
        all_add_list = add_desc.index[1:]
        if len(drop_adds) == 0:
            use_add_list = all_add_list
        else:
            use_add_list = list(set(all_add_list) - set(drop_adds))
    else:
        all_add_list, use_add_list = [], []

    if supp_num != 0:
        all_supp_list = supp_desc.index
        if len(drop_supp_list) == 0:
            use_supp_list = all_supp_list
        else:
            use_supp_list = list(set(all_supp_list) - set(drop_supp_list))
        if len(essential_supp) != 0:
            for i in range(len(essential_supp)):
                essential_supp[i] = supp_name_dict[essential_supp[i]]
            use_supp_list = essential_supp
    else:
        all_supp_list, use_supp_list = [], []

    elem_lists_dict = {
        'all_pgm': all_pgm_list, 'use_pgm': use_pgm_list,
        'all_add': all_add_list, 'use_add': use_add_list,
        'all_supp': all_supp_list, 'use_supp': use_supp_list,
        'CalT': CalT_list
        }
    return elem_lists_dict




# 探索する触媒組成に応じたdictを作成(後の関数内で呼び出す)
def cols_order(condition: dict) -> dict:

    cols_order = {}
    cand_pgm_num = condition['cand_pgm_num']
    cand_add_num = condition['cand_add_num']
    cand_supp_num = condition['cand_supp_num']
    CalT_num = condition['CalT_num']

    if cand_pgm_num != 0 and cand_add_num != 0 \
        and cand_supp_num != 0 and CalT_num != 0:
        cols_order = {'pgm': 0, 'add': 1, 'supp': 2, 'CalT': 3}

    elif cand_pgm_num != 0 and cand_add_num != 0 \
        and cand_supp_num != 0 and CalT_num == 0:
            cols_order = {'pgm': 0, 'add': 1, 'supp': 2}

    elif cand_pgm_num == 0 and cand_add_num != 0 \
        and cand_supp_num != 0 and CalT_num != 0:
        cols_order = {'add': 0, 'supp': 1, 'CalT': 2}

    elif cand_pgm_num == 0 and cand_add_num != 0 \
        and cand_supp_num == 0 and CalT_num == 0:
        cols_order = {'add': 0}

    return cols_order



# 実験データを読み込む関数
def read_data(condition: dict, desc_cols: list,
              read_sheet_name: str = 'data_sheet_name', idx: int = None) -> pd.DataFrame:

    date, Reaction = condition['date'], condition['Reaction']
    data_sheet_name, skip_rows = condition[read_sheet_name], condition['skip_rows']
    supp_num = condition['supp_num']
    supp_name_dict = desc_cols['supp_name_dict']
    data_file_name = f'data/{date}_{Reaction}_dataset.xlsx'

    data = pd.read_excel(data_file_name, sheet_name=data_sheet_name, skiprows=skip_rows)

    data_cols = setting.data_columns(condition)
    elem_cols, wt_cols = data_cols['elem'], data_cols['wt']

    data.loc[:, elem_cols] = data.loc[:, elem_cols].fillna('H')
    data.loc[:, wt_cols] = data.loc[:, wt_cols].fillna(0)

    if supp_num != 0:
        for i in range(len(data)):
            data.loc[i, 'Support_name'] = supp_name_dict[data.loc[i, 'Support_name']]

    if idx is not None:
        #print(Reaction, 'Iteration =', idx)
        data_idx = data[data['Iteration'] <= idx]
        data_idx = data_idx.reset_index(drop=True)
        data = data_idx

    return data



#　Descriptorファイルを読み込む
def read_desc(condition: dict, desc_cols: list, elem: str = 'all',
              select_desc: bool = True, local: bool = False, standard: bool = False) -> tuple:

    Reaction = condition['Reaction']
    pgm_num = condition['pgm_num']
    add_num = condition['add_num']
    supp_num = condition['supp_num']

    elem_comp_list = [pgm_num, add_num, supp_num]
    desc_file_name = 'data/Descriptors_alcohol.xlsx'

    if elem == 'pgm' or elem == 'add' or elem == 'all':
        desc_sheet_name = 'Descriptors_elem'
        use_desc = desc_cols['basic_desc_columns']
        elem_desc_dict = desc_cols['elem_desc_dict']
        for i, j in enumerate(use_desc):
            if j in list(elem_desc_dict.keys()):
                use_desc[i] = elem_desc_dict[j]
        desc = pd.read_excel(desc_file_name, sheet_name=desc_sheet_name,
                             index_col='Symbol')
        desc = desc.rename(columns=elem_desc_dict)

    elif elem == 'supp':
        desc_sheet_name = 'Descriptors_supp'
        use_desc = desc_cols['all_supp_desc']
        supp_desc_dict = desc_cols['supp_desc_dict']
        for i, j in enumerate(use_desc):
            if j in list(supp_desc_dict.keys()):
                use_desc[i] = supp_desc_dict[j]
        desc = pd.read_excel(desc_file_name, sheet_name=desc_sheet_name,
                             index_col='Support_name_python')
        desc = desc.rename(columns=supp_desc_dict)

    all_elems = list(desc.index)
    noble_gas = desc_cols['noble_gas']
    drop_elems = desc_cols['drop_elems']

    if elem == 'pgm':
        use_elem = desc_cols['pgm_plus_ReAu']
        desc = desc.loc[use_elem].drop(['Re', 'Os'], axis=0)
        desc = desc[use_desc]
        desc = desc.fillna(desc.mean())
        if len(elem_comp_list) - elem_comp_list.count(0) != 1:
            desc = desc.rename(columns=lambda s: s+' [PGMs]')
            aw = desc.loc[:, 'AW [PGMs]']
        else:
            aw = desc.loc[:, 'AW']

    elif elem == 'add':
        use_elem = [i for i in all_elems if i not in noble_gas+drop_elems]
        desc = desc.loc[use_elem]
        desc = desc[use_desc]
        desc = desc.fillna(desc.mean())
        if len(elem_comp_list) - elem_comp_list.count(0) != 1:
            desc = desc.rename(columns=lambda s: s+' [Additives]')
            aw = desc.loc[:, 'AW [Additives]']
        else:
            aw = desc.loc[:, 'AW']

    elif elem == 'supp':
        use_elem = desc.index[desc[Reaction] == 1]
        desc = desc.loc[use_elem]
        #desc = desc.set_index('Support_name_python', inplace=True)
        desc = desc[use_desc].replace('-', 0)
        desc = desc[use_desc]
        desc = desc.fillna(desc.mean())
        if len(elem_comp_list) - elem_comp_list.count(0) != 1:
            desc = desc.rename(columns=lambda s: s+' [Supports]')
            aw = desc.loc[:, 'MW [Supports]']
        else:
            aw = desc.loc[:, 'MW']

    elif elem == 'all':
        desc = desc[use_desc]
        desc = desc.fillna(desc.mean())
        aw = desc.loc[:, 'AW']

    if local == False:
        selec_desc = conditions.desc_select(Reaction)

    ## Remove the '#' in the Jupyter Notebook version.
    # else:
        #selec_desc = desc_select(Reaction)

    if elem == 'pgm' or elem == 'add':
        for i, j in enumerate(selec_desc['use_'+elem]):
            if j in list(elem_desc_dict.keys()):
                selec_desc['use_'+elem][i] = elem_desc_dict[j]
        for i, j in enumerate(selec_desc['drop_'+elem]):
            if j in list(elem_desc_dict.keys()):
                selec_desc['drop_'+elem][i] = elem_desc_dict[j]
    elif elem == 'supp':
        for i, j in enumerate(selec_desc['use_'+elem]):
            if j in list(supp_desc_dict.keys()):
                selec_desc['use_'+elem][i] = supp_desc_dict[j]
        for i, j in enumerate(selec_desc['drop_'+elem]):
            if j in list(supp_desc_dict.keys()):
                selec_desc['drop_'+elem][i] = supp_desc_dict[j]

    if select_desc == True:
        if elem == 'pgm':
            selec_desc['use_'+elem] = [s +' [PGMs]' for s in selec_desc['use_'+elem]]
        if elem == 'add':
            selec_desc['use_'+elem] = [s +' [Additives]' for s in selec_desc['use_'+elem]]
        if elem == 'supp':
            selec_desc['use_'+elem] = [s +' [Supports]' for s in selec_desc['use_'+elem]]
        if elem != 'all':
            desc = desc.loc[:, selec_desc['use_'+elem]]
    else:
        if elem == 'pgm':
            selec_desc['drop_'+elem] = [s +' [PGMs]' for s in selec_desc['drop_'+elem]]
        if elem == 'add':
            selec_desc['drop_'+elem] = [s +' [Additives]' for s in selec_desc['drop_'+elem]]
        if elem == 'supp':
            selec_desc['drop_'+elem] = [s +' [Supports]' for s in selec_desc['drop_'+elem]]
        if elem != 'all':
            desc = desc.drop(selec_desc['drop_'+elem], axis=1)
    desc = desc.dropna(how='any', axis=1)

    if standard == True:
        desc = (desc - desc.mean()) / desc.std(ddof=1)
    return desc, aw




# 触媒候補データ(cand_sum)を読み込む関数
def read_cand(condition: dict) -> pd.DataFrame:

    date, Reaction = condition['date'], condition['Reaction']
    ML_model, Search_method = condition['ML_model'], condition['Search_method']

    pgm_num = condition['pgm_num']
    add_num = condition['add_num']
    supp_num = condition['supp_num']

    pgm_model = condition['pgm_model']
    add_model = condition['add_model']
    supp_model = condition['supp_model']

    # results_csv_file_all
    if pgm_num != 0 and add_num != 0 and supp_num != 0:
        cand_file_name = \
            f'results/{date}_{Reaction}_cand_sum_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}.csv'


    data_cols = setting.data_columns(condition)
    cand_elem, cand_wt = data_cols['cand_elem'], data_cols['cand_wt']

    cand = cand_separation(cand_file_name, condition)
    cand = cand_str(cand, condition, data_cols)
    cand.drop_duplicates(subset=['Top catal.'], inplace=True)

    cand.loc[:, cand_elem] = cand.loc[:, cand_elem].fillna('H')
    cand.loc[:, cand_wt] = cand.loc[:, cand_wt].fillna(0)
    cand = cand.reset_index()
    return cand



# Feat.(X) を定義する関数
def make_feat(elem: str, elem_num: int, model: object,
              data: pd.DataFrame, desc: pd.DataFrame, aw: pd.Series) -> tuple:

    if elem == 'PGM' or elem == 'Ad':
        if elem_num != 0:
            feat_comp = pd.DataFrame(
                index=np.arange(len(data)), columns=desc.index)
            for i in range(len(data)):
                for j in range(elem_num):
                    feat_comp.loc[i, data.loc[i, f'{elem}{j+1}']] = data.loc[i, f'{elem}{j+1}_wt%']
            feat_comp = feat_comp.fillna(0)
            if 'H' in desc.index:
                feat_comp = feat_comp.drop('H', axis=1)
                feat_comp_cols = list(desc.index[1:])
            else:
                feat_comp_cols = list(desc.index)

            feat_desc = sum([
                np.multiply(np.array(desc.loc[data.loc[:, f"{elem}{i+1}"]]),
                            np.array(data[f"{elem}{i+1}_wt%"]).reshape(-1, 1)) /
                np.array(aw.loc[data.loc[:, f"{elem}{i+1}"]]).reshape(-1, 1) for i in range(elem_num)
            ])
            feat_desc_cols = list(desc.columns)

            if model == 0:
                feat = feat_comp
                feat_cols = feat_comp_cols
            elif model == 1:
                feat = np.hstack((feat_comp, feat_desc))
                feat_cols = list(feat_comp_cols) + list(feat_desc_cols)
            elif model == 2:
                feat = feat_desc
                feat_cols = feat_desc_cols
        else:
            feat, feat_cols = None, None

    elif elem == 'Supp':
        if elem_num != 0:
            feat_comp = pd.DataFrame(index=np.arange(len(data)), columns=desc.index)
            for i in range(len(data)):
                feat_comp.loc[i, data.loc[i, 'Support_name']] = 1
            feat_comp = feat_comp.fillna(0)
            feat_comp_cols = list(desc.index)

            feat_desc = desc.loc[data.loc[:, 'Support_name']]
            feat_desc = feat_desc.replace('-', 0)
            feat_desc_cols = list(desc.columns)

            if model == 0:
                feat = feat_comp
                feat_cols = feat_comp_cols
            elif model == 1:
                feat = np.hstack((feat_comp, feat_desc))
                feat_cols = list(feat_comp_cols) + list(feat_desc_cols)
            elif model == 2:
                feat = feat_desc
                feat_cols = feat_desc_cols
        else:
            feat, feat_cols = None, None

    elif elem == 'Calc T':
        if elem_num != 0:
            feat = np.array(data.loc[:, elem]).reshape(-1, 1)
            feat_cols = [elem]
        else:
            feat, feat_cols = None, None
    return feat, feat_cols



# Targetを定義する関数
def make_target(Reaction: str, data: pd.DataFrame, name: str) -> pd.Series:

    if Reaction == 'CH3OH_Batch':
        target = data.loc[:, name]
        target = target.fillna(target.mean())
        target = target.astype(float)
    return target




# Sorted Weighted Elemental Descriptor (SWED) 表現を定義する関数
# 参考文献 (ChemCatChem, 2021, 13, 3636-3655.)
# URL: https://chemistry-europe.onlinelibrary.wiley.com/doi/full/10.1002/cctc.202100495
def make_swed(elem: str, elem_num: int, data: pd.DataFrame, desc: pd.DataFrame,
              aw: pd.Series, data_cols: list) -> pd.DataFrame:
    """
    Define the Sorted Weighted Elemental Descriptor (SWED) representation for elements.

    Parameters
    ----------
    elem : str
        The type of element to process. Can be 'PGM' (PGM elements), 'Promoter', 'Ad' (additives), or 'Supp' (supports).
    elem_num : int
        The number of elements to process for the given type.
    data : pd.DataFrame
        The DataFrame containing the element composition and weight percentage data.
    desc : pd.DataFrame
        The DataFrame containing descriptor data for the elements.
    aw : pd.Series
        The atomic weights (or molecular weights for supports) of the elements.
    data_cols : list
        A list of column names or mappings used to access specific columns in `data`, particularly for element weights.

    Returns
    -------
    pd.DataFrame
        A DataFrame representing the SWED (Sorted Weighted Elemental Descriptor) values for each element and descriptor.
        If `elem_num` is zero, returns `None`.

    Notes
    -----
    - The SWED (Sorted Weighted Elemental Descriptor) values are calculated by multiplying elemental descriptors with their respective weight percentages and dividing by their atomic weights.
    - For non-support elements ('PGM' or 'Ad'), the calculation is performed element-wise, while for supports ('Supp'), the remaining weight percentage (100 - sum of element weights) is used.
    - Based on the methodology in the reference: ChemCatChem, 2021, 13, 3636-3655.
    - URL: https://chemistry-europe.onlinelibrary.wiley.com/doi/full/10.1002/cctc.202100495
    """
    if elem_num != 0:
        elem_swed_names = []
        for i in range(elem_num):
            for j in desc.columns:
                elem_swed_names.append(f'{j} [{i+1}]')

        swed_elem = pd.DataFrame(columns=elem_swed_names)
        if elem != 'Supp':
            for i in range(len(data)):
                swed_elem.loc[i] = np.array(
                    list(
                        flatten(
                            [(
                                np.array(desc.loc[data.loc[i, f"{elem}{j+1}"]]) *
                                np.array(data.loc[i, f"{elem}{j+1}_wt%"]) /
                                np.array(aw.loc[data.loc[i, f"{elem}{j+1}"]])
                            ) for j in range(elem_num)]
                        )))
        else:
            elem_wt_sum = data.loc[:, data_cols['wt']].sum(axis=1)
            supp_wt = 100 - elem_wt_sum
            for i in range(len(data)):
                swed_elem.loc[i] = np.array(desc.loc[data.loc[i, 'Support_name']]) * \
                    np.array(supp_wt[i]) / \
                    np.array(aw.loc[data.loc[i, 'Support_name']])
    else:
        swed_elem = None
    return swed_elem



# datasetと DescriptorのExcelファイルを読み込み、feat, target等必要な形状に加工する。
def data_convert(condition: dict, data_cols: list, desc_cols: list, sheet_name: str = 'data_sheet_name',
                 select_desc: bool = True, use_models: list = [],
                 local: bool = False, idx: int = None) -> dict:

    converted = {}

    data = read_data(condition, desc_cols, read_sheet_name=sheet_name, idx=idx)
    
    pgm_desc, pgm_aw = read_desc(condition=condition, desc_cols=desc_cols, elem='pgm',
                                 select_desc=select_desc, local=local, standard=condition['desc_std'])
    
    add_desc, add_aw = read_desc(condition=condition, desc_cols=desc_cols, elem='add',
                                 select_desc=select_desc, local=local, standard=condition['desc_std'])
    
    supp_desc, supp_mw = read_desc(condition=condition, desc_cols=desc_cols, elem='supp',
                                   select_desc=select_desc, local=local, standard=condition['desc_std'])

    if len(use_models) == 0:
        pgm_model = condition['pgm_model']
        add_model = condition['add_model']
        supp_model = condition['supp_model']

    elif len(use_models) == 1:
        add_model = use_models[0]
        pgm_model, supp_model = 2, 2

    elif len(use_models) == 2:
        add_model, supp_model = use_models[0], use_models[1]
        pgm_model = 2

    elif len(use_models) == 3:
        pgm_model, add_model, supp_model = use_models[0], use_models[1], use_models[2]

    feat_pgm, feat_pgm_cols = make_feat(elem='PGM', elem_num=condition['pgm_num'],
                                        model=pgm_model, data=data, desc=pgm_desc, aw=pgm_aw)
    
    feat_add, feat_add_cols = make_feat(elem='Ad', elem_num=condition['add_num'], 
                                        model=add_model, data=data, desc=add_desc, aw=add_aw)
    
    feat_supp, feat_supp_cols = make_feat(elem='Supp', elem_num=condition['supp_num'],
                                          model=supp_model, data=data, desc=supp_desc, aw=supp_mw)

    feat_CalT, feat_CalT_cols = make_feat(elem='Calc T', elem_num=condition['CalT_num'],
                                            model=0, data=data, desc=add_desc, aw=add_aw)


    feat_list = [feat_pgm, feat_add, feat_supp, feat_CalT]
    feat_cols_list = [feat_pgm_cols,feat_add_cols, feat_supp_cols, feat_CalT_cols]
    
    feat_list = [i for i in feat_list if i is not None]
    feat = np.hstack(feat_list)

    feat_cols_list = [i for i in feat_cols_list if i is not None]
    feat_cols = list(itertools.chain.from_iterable(feat_cols_list))

    feat = pd.DataFrame(feat, columns=feat_cols)

    swed_pgm = make_swed(elem='PGM', elem_num=condition['pgm_num'],
                         data=data, desc=pgm_desc, aw=pgm_aw, data_cols=data_cols)
    
    swed_add = make_swed(elem='Ad', elem_num=condition['add_num'],
                         data=data, desc=add_desc, aw=add_aw, data_cols=data_cols)
    
    swed_supp = make_swed(elem='Supp', elem_num=condition['supp_num'],
                          data=data, desc=supp_desc, aw=supp_mw, data_cols=data_cols)

    target = make_target(Reaction=condition['Reaction'], data=data, name=condition['target_name'])

    converted = {
        'data': data,
        'pgm_desc': pgm_desc, 'add_desc': add_desc, 'supp_desc': supp_desc,
        'pgm_aw': pgm_aw, 'add_aw': add_aw, 'supp_mw': supp_mw,
        'swed_pgm': swed_pgm, 'swed_add': swed_add, 'swed_supp': swed_supp,
        'feat': feat, 'target': target
    }
    return converted




# datasetと DescriptorのExcelファイルを読み込み、feat, target等必要な形状に加工する。
def cand_convert(condition: dict, converted: dict, data_cols: list, desc_cols: list,
                 select_desc: bool = True, use_models: list = [], local: bool = False) -> dict:
    
    cand_dict = {}
    
    data = converted['data']
    data = data_str(data, condition, data_cols)

    cand = read_cand(condition)
    dupl_cat = list(set(data['Catalyst']) & set(cand['Top catal.']))
    if len(dupl_cat) != 0:
        for i in range(len(dupl_cat)):
            cand.drop(list(cand[cand['Top catal.'] == dupl_cat[i]].index), inplace=True)
    cand.reset_index(drop=True, inplace=True)

    for i in list(cand.columns[:-1]): # drop 'Top catal.'
        if type(cand.loc[0, i]) == str and ',' in cand.loc[0, i]:
            cand[i] = cand[i].str.strip(',')
            if i != 'Support_name':
                cand[i] = cand[i].astype(float)
    
    pgm_desc, pgm_aw = read_desc(condition=condition, desc_cols=desc_cols, elem='pgm',
                                 select_desc=select_desc, local=local, standard=condition['desc_std'])
      
    add_desc, add_aw = read_desc(condition=condition, desc_cols=desc_cols, elem='add',
                                 select_desc=select_desc, local=local, standard=condition['desc_std'])
    
    supp_desc, supp_mw = read_desc(condition=condition, desc_cols=desc_cols, elem='supp',
                                   select_desc=select_desc, local=local, standard=condition['desc_std'])

    if len(use_models) == 0:
        pgm_model = condition['pgm_model']
        add_model = condition['add_model']
        supp_model = condition['supp_model']

    elif len(use_models) == 1:
        add_model = use_models[0]
        pgm_model, supp_model = 2, 2

    elif len(use_models) == 2:
        add_model, supp_model = use_models[0], use_models[1]
        pgm_model = 2

    elif len(use_models) == 3:
        pgm_model, add_model, supp_model = use_models[0], use_models[1], use_models[2]


    feat_pgm, feat_pgm_cols = make_feat(elem='PGM', elem_num=condition['cand_pgm_num'],
                                        model=pgm_model, data=cand, desc=pgm_desc, aw=pgm_aw) # <-- num=cand_nun, data=cand !!!
    
    feat_add, feat_add_cols = make_feat(elem='Ad', elem_num=condition['cand_add_num'], 
                                        model=add_model, data=cand, desc=add_desc, aw=add_aw) # <-- num=cand_nun, data=cand !!!
    
    feat_supp, feat_supp_cols = make_feat(elem='Supp', elem_num=condition['cand_supp_num'],
                                          model=supp_model, data=cand, desc=supp_desc, aw=supp_mw) # <-- num=cand_nun, data=cand !!!

    feat_CalT, feat_CalT_cols = make_feat(elem='Calc T', elem_num=condition['CalT_num'],
                                          model=0, data=cand, desc=add_desc, aw=add_aw) # <-- data=cand !!!

    feat_list = [feat_pgm, feat_add, feat_supp, feat_CalT]
    
    feat_cols_list = [feat_pgm_cols, feat_add_cols, feat_supp_cols, feat_CalT_cols]
    
    feat_list = [i for i in feat_list if i is not None]
    feat = np.hstack(feat_list)
    feat_cols_list = [i for i in feat_cols_list if i is not None]
    feat_cols = list(itertools.chain.from_iterable(feat_cols_list))
    feat = pd.DataFrame(feat, columns=feat_cols)
    target = cand.loc[:, 'ei']

    cand_dict = {
        'cand': cand,
        'pgm_desc': pgm_desc, 'add_desc': add_desc, 'supp_desc': supp_desc,
        'pgm_aw': pgm_aw, 'add_aw': add_aw, 'supp_mw': supp_mw,
        'feat': feat, 'target': target
        }
    return cand_dict




# 各元素成分ごとに、wtの組み合わせを作る関数
def elem_wt_comb(elem_wt: list, ess_elem_wt: list, 
                 cand_elem_num: int, fix_elem_num: int, max_elem_wt: float) -> iter:
    """
    Generate combinations of elemental weights for given element components.

    Parameters
    ----------
    elem_wt : list
        A list of candidate elemental weights to combine.
    ess_elem_wt : list
        A list of fixed essential element weights.
    cand_elem_num : int
        The total number of candidate elements.
    fix_elem_num : int
        The number of fixed essential elements whose weights will not change.
    max_elem_wt : float
        The maximum allowable sum for the combined elemental weights. If `None`, no limit is applied.

    Returns
    -------
    iter
        An iterator over the combinations of elemental weights. Each combination is a list of weights for all elements.
        If `cand_elem_num` is zero, returns `None`.

    Notes
    -----
    - The function generates all possible combinations of weights for the candidate elements using the `itertools.product` method.
    - If fixed essential elements are provided (`fix_elem_num` > 0), their weights are included in the combinations.
    - If `max_elem_wt` is provided, combinations whose total weight exceeds this value are excluded.
    """
    if cand_elem_num != 0:
        elem_wt_comb_full = itertools.product(elem_wt, repeat=cand_elem_num-fix_elem_num)
        if fix_elem_num != 0:
            elem_wt_comb_fix = itertools.product(ess_elem_wt, repeat=fix_elem_num)
            elem_wt_comb = itertools.product(elem_wt_comb_fix, elem_wt_comb_full)
        else:
            elem_wt_comb = itertools.product(elem_wt, repeat=cand_elem_num)

        elem_wt_comb = np.array(list(flatten(elem_wt_comb))).reshape(-1, cand_elem_num)
        if max_elem_wt is not None:
            elem_wt_comb = elem_wt_comb[elem_wt_comb.sum(axis=1) < max_elem_wt]
        elem_wt_comb = iter(elem_wt_comb)
    else:
        elem_wt_comb = None
    return elem_wt_comb



# 上で定義したelem_wt_comb関数を成分の数だけ結合し、wtの全ての組み合わせを作成する。
def all_wt_comb(condition: dict) -> tuple:

    pgm_wt = condition['pgm_wt']
    add_wt = condition['add_wt']

    ess_pgm_wt = condition['essential_pgm_wt']
    ess_add_wt = condition['essential_add_wt']

    cand_pgm_num = condition['cand_pgm_num']
    cand_add_num = condition['cand_add_num']
    cand_supp_num = condition['cand_supp_num']

    fix_pgm_num = condition['fix_pgm_num']
    fix_add_num = condition['fix_add_num']

    max_pgm_wt = condition['max_pgm_wt']
    max_add_wt = condition['max_add_wt']

    elem_wt_pgm = elem_wt_comb(pgm_wt, ess_pgm_wt, cand_pgm_num,
                               fix_pgm_num, max_pgm_wt)
    
    elem_wt_add = elem_wt_comb(add_wt, ess_add_wt, cand_add_num,
                               fix_add_num, max_add_wt)

    _lis_wt = [elem_wt_pgm, elem_wt_add]
    comb_wt = [z for z in _lis_wt if z]
    all_comb_wt = itertools.product(*comb_wt)
    contens = len(_lis_wt) - _lis_wt.count(None)

    if contens == 1:
        elem_wt = [x1[0] for x1 in all_comb_wt]

    elif contens == 2:
        elem_wt = [(*x1, *x2) for (x1, x2) in all_comb_wt]

    elif contens == 3:
        elem_wt = [(*x1, *x2, *x3) for (x1, x2, x3) in all_comb_wt]

    if cand_supp_num != 0:
        # elem_wt = [elem_wt[i] + tuple([100-sum(elem_wt[i])]) for i in range(len(elem_wt))]
        elem_wt = [np.append(arr, 100 - np.sum(arr)) for arr in elem_wt]

    num_elem_wt = len(elem_wt)
    return elem_wt, num_elem_wt




# 必須(essential)元素が登場するリストとその他の元素(未登場の必須元素も含まれ得る)が登場するリストをそれぞれ作成。
def gen_fix_and_res(use_elem_list: list, essential_elems: list,
                    cand_elem_num: int, fix_elem_num: int, duplicate: bool = True) -> tuple:
    """
    Generate two lists: one containing combinations of essential elements (fixed elements) 
    and the other containing combinations of remaining elements (residual elements), which 
    may include unused essential elements.

    Parameters
    ----------
    use_elem_list : list
        The list of all elements to be considered for combination.
    essential_elems : list
        The list of essential elements that must be included in the fixed element combinations.
    cand_elem_num : int
        The total number of candidate elements for the combination.
    fix_elem_num : int
        The number of essential elements to be included in the fixed element list.
    duplicate : bool, optional
        If True, allows duplicates between the essential elements and the residual elements, by default True.

    Returns
    -------
    tuple
        A tuple containing:
        - elem_comb_fix (list of tuples): Combinations of the essential elements (fixed elements).
        - elem_comb_res (list of tuples): Combinations of the remaining elements (residual elements). 
          If `duplicate` is True, the remaining elements can include essential elements.

    Notes
    -----
    - The function generates combinations of fixed elements using `itertools.combinations`.
    - The residual elements are also combined using `itertools.combinations`, with the option to include essential elements 
      if `duplicate` is set to True.
    - The residual elements list will exclude essential elements if `duplicate` is False.
    """
    # residue_elems: 必須元素を除いた残りの元素リストを作成
    residue_elems = list(set(use_elem_list) - set(essential_elems))
    
    # elem_comb_fix: 固定される必須元素の組み合わせを生成
    elem_comb_fix = list(itertools.combinations(essential_elems, fix_elem_num))
    
    # elem_comb_res: 残りの元素の組み合わせを生成 (重複を許すかどうかを指定)
    if duplicate == True:
        # 重複を許す場合、全ての元素リストから残りの組み合わせを生成
        elem_comb_res = list(itertools.combinations(use_elem_list, cand_elem_num - fix_elem_num))
    else:
        # 重複を許さない場合、必須元素を除いた残りの元素リストから組み合わせを生成
        elem_comb_res = list(itertools.combinations(residue_elems, cand_elem_num - fix_elem_num))
    
    return elem_comb_fix, elem_comb_res



# 上のgen_fix_and_res関数で作成した必須元素リストと、その他元素リストを結合する。
def fix_and_res_comb(fix_list: list, res_list: list) -> list:
    """
    Combine the fixed element list and the residual element list into all possible combinations.

    Parameters
    ----------
    fix_list : list
        A list containing combinations of fixed elements (essential elements).
    res_list : list
        A list containing combinations of residual elements (non-essential elements).

    Returns
    -------
    list
        A list of tuples, where each tuple is a combination of elements from both the fixed and residual lists.
    """
    _lis = [fix_list, res_list]
    comb = [z for z in _lis if z is not None]
    comb_ = itertools.product(*comb)
    comb_list = [(*x1, *x2) for (x1, x2) in comb_]
    return comb_list



# 上のfix_and_res_comb関数で作成したリストに含まれる重複(e.g. ['Pt', 'Pt'])成分を落とす。
def drop_duplicate(elem_comb: list, fix_elem_num: int) -> list:
    """
    Remove combinations from the list that contain duplicate elements across the fixed and residual parts.

    Parameters
    ----------
    elem_comb : list
        A list of element combinations. Each combination is a tuple containing both fixed and residual elements.
    fix_elem_num : int
        The number of fixed elements in each combination.

    Returns
    -------
    list
        A list of element combinations where no element appears in both the fixed and residual parts.

    Notes
    -----
    - This function checks for duplicates between the fixed part (`elem_comb[i][:fix_elem_num]`) 
      and the residual part (`elem_comb[i][fix_elem_num:]`) of each combination.
    - If no duplicates are found, the combination is added to the returned list.
    """
    elem_comb_dup = []
    for i in range(len(elem_comb)):
        if len(list(set(elem_comb[i][:fix_elem_num]) & set(elem_comb[i][fix_elem_num:]))) == 0:
            elem_comb_dup.append(elem_comb[i])
    return elem_comb_dup



# drop_duplicate関数で作成した重複を含まないリストを要素ごとに分解する。
def make_elem_comp(elem_comb: list, cand_elem_num: int) -> dict:
    """
    Decompose a list of element combinations into a dictionary where each key corresponds to a position in the combination,
    and the value is a list of elements at that position.

    Parameters
    ----------
    elem_comb : list
        A list of element combinations, where each combination is a tuple of elements.
    cand_elem_num : int
        The number of candidate elements in each combination.

    Returns
    -------
    dict
        A dictionary where the keys are the index positions (0 to `cand_elem_num` - 1) and the values 
        are lists of elements that appear in those positions across all combinations.

    Notes
    -----
    - This function iterates over each combination and decomposes the elements by their position into separate lists.
    - The result is a dictionary where each key corresponds to a specific element position in the combination, 
      and the value is a list of all elements that appear in that position.
    """
    elem_comp_dict = {}
    for i in range(cand_elem_num):
        elem_comp_dict[i] = [x[i] for x in elem_comb]
    return elem_comp_dict



# 触媒組成 (+ 実験条件) の全組み合わせを計算する。
def catal_comb(condition: dict, elem_lists: list) -> tuple:

    cand_pgm_num = condition['cand_pgm_num']
    cand_add_num = condition['cand_add_num']
    cand_supp_num = condition['cand_supp_num']

    CalT_num = condition['CalT_num']

    fix_pgm_num = condition['fix_pgm_num']
    fix_add_num = condition['fix_add_num']

    essential_pgms = condition['essential_pgms']
    essential_adds = condition['essential_adds']

    use_pgm_list = elem_lists['use_pgm']
    use_add_list = elem_lists['use_add']
    use_supp_list = elem_lists['use_supp']
    CalT_list = elem_lists['CalT']

    if cand_pgm_num != 0:
        if fix_pgm_num != 0:
            elem_comb_pgm_fix, elem_comb_pgm_res = \
                gen_fix_and_res(use_pgm_list, essential_pgms,
                                cand_pgm_num, fix_pgm_num, duplicate=True)
            elem_comb_pgm = fix_and_res_comb(elem_comb_pgm_fix, elem_comb_pgm_res)
            del elem_comb_pgm_fix
            del elem_comb_pgm_res
            elem_comb_pgm = drop_duplicate(elem_comb_pgm, fix_pgm_num)
        else:
            elem_comb_pgm = itertools.combinations(use_pgm_list, cand_pgm_num)
            elem_comb_pgm = list(elem_comb_pgm)
        comb_pgm_num = len(elem_comb_pgm)
    else:
        elem_comb_pgm = None
        comb_pgm_num = 1


    if cand_add_num != 0:
        if fix_add_num != 0:
            elem_comb_add_fix, elem_comb_add_res \
                = gen_fix_and_res(use_add_list, essential_adds,
                                  cand_add_num, fix_add_num, duplicate=True)
            elem_comb_add = fix_and_res_comb(elem_comb_add_fix, elem_comb_add_res)
            del elem_comb_add_fix
            del elem_comb_add_res
            elem_comb_add = drop_duplicate(elem_comb_add, fix_add_num)
        else:
            elem_comb_add = itertools.combinations(use_add_list, cand_add_num)
            elem_comb_add = list(elem_comb_add)
        comb_add_num = len(elem_comb_add)
    else:
        elem_comb_add = None
        comb_add_num = 1


    if cand_supp_num != 0:
        elem_comb_supp = use_supp_list
        comb_supp_num = len(elem_comb_supp)
    else:
        elem_comb_supp = None
        comb_supp_num = 1


    if CalT_num != 0:
        elem_comb_CalT = CalT_list
        comb_CalT_num = len(elem_comb_CalT)
    else:
        elem_comb_CalT = None
        comb_CalT_num = 1


    total_comb_num = comb_pgm_num * comb_add_num \
        * comb_supp_num * comb_CalT_num

    _lis = [
        elem_comb_pgm, elem_comb_add,
        elem_comb_supp, elem_comb_CalT
        ]
    
    comb = [z for z in _lis if z is not None]
    all_comb = itertools.product(*comb)

    return all_comb, total_comb_num




# 触媒組成を one-hot表現に変換する関数
def one_hot(read_col_name: str, elem_list: list, cand_elem_num: int, cand_cols_order: dict,
            condition: dict, elem_lists: list, slice_size: int, start: int, end: int) -> list:
    """
    Convert catalyst composition into a one-hot encoded representation.

    Parameters
    ----------
    read_col_name : str
        The name of the column to read from the catalyst combinations.
    elem_list : list
        The list of elements to be used for one-hot encoding.
    cand_elem_num : int
        The number of candidate elements to be used for the one-hot encoding.
    cand_cols_order : dict
        A dictionary specifying the column order for candidate elements.
    condition : dict
        A dictionary containing various parameters for the catalyst and experimental conditions.
    elem_lists : list
        A list containing the usable elements for PGM, additives, and supports.
    slice_size : int
        The number of rows (combinations) to process at once (batch size).
    start : int
        The starting index for slicing the combinations.
    end : int
        The ending index for slicing the combinations.

    Returns
    -------
    list
        A list of compressed sparse row (CSR) matrices, each representing the one-hot encoded 
        matrix for the corresponding element positions.

    Notes
    -----
    - If `cand_elem_num` is 1, the function directly converts the single element combination to one-hot encoding.
    - If `cand_elem_num` is greater than 1, it processes each element position separately.
    - Uses `catal_comb` to generate the combinations of catalyst elements and slices them based on the `start` and `end` indices.
    - The result is a list of one-hot encoded matrices (CSR format) for each element position.
    """
    elem_comp_dict = {}
    elem_cols = {}
    ones = np.ones(slice_size)
    rows = np.arange(0, slice_size, 1)
    int_code = {x: j for (j, x) in enumerate(elem_list)}
    read_col = cand_cols_order[read_col_name]

    if cand_elem_num == 1:
        all_comb, total_comb_num = catal_comb(condition, elem_lists)
        all_comb_slice = itertools.islice(all_comb, start, end)
        elem_comp_dict = [u[read_col] for u in all_comb_slice]
        elem_cols = np.asarray([int_code[v] for v in elem_comp_dict])
        elem_oh = [sparse.csr_matrix((ones, (rows, elem_cols)),
                    shape=(slice_size, len(elem_list)))]
    else:
        for n in range(cand_elem_num):
            all_comb, total_comb_num = catal_comb(condition, elem_lists)
            all_comb_slice = itertools.islice(all_comb, start, end)
            elem_comp_dict[n] = [u[read_col][n] for u in all_comb_slice]
            elem_cols[n] = np.asarray([int_code[v] for v in elem_comp_dict[n]])
        elem_oh = [sparse.csr_matrix((ones, (rows, elem_cols[n])),
                    shape=(slice_size, len(elem_list))) for n in range(cand_elem_num)]
    return elem_oh



# 実験条件を入力表現に変換する関数
def exp_rep(read_col_name: str, cand_cols_order: dict, condition: dict,
            elem_lists: list, start: int, end: int) -> sparse.csr_matrix:
    """
    Convert experimental conditions into an input representation.

    Parameters
    ----------
    read_col_name : str
        The name of the column to read from the catalyst combinations.
    cand_cols_order : dict
        A dictionary specifying the column order for candidate elements and experimental conditions.
    condition : dict
        A dictionary containing various parameters for the catalyst and experimental conditions.
    elem_lists : list
        A list containing the usable elements for PGM, additives, and supports.
    start : int
        The starting index for slicing the combinations.
    end : int
        The ending index for slicing the combinations.

    Returns
    -------
    sparse.csr_matrix
        A sparse CSR matrix representing the experimental condition input for the specified range.

    Notes
    -----
    - This function generates all possible combinations of catalyst compositions using the `catal_comb` function.
    - It slices the generated combinations based on the `start` and `end` indices, and extracts the experimental condition specified by `read_col_name`.
    - The experimental conditions are converted into a sparse CSR matrix.
    """
    read_col = cand_cols_order[read_col_name]
    all_comb, total_comb_num = catal_comb(condition, elem_lists)
    all_comb_slice = itertools.islice(all_comb, start, end)
    exp_dict = [u[read_col] for u in all_comb_slice]
    exp_col = np.array(exp_dict).reshape(-1, 1)
    exp_rep = sparse.csr_matrix(exp_col)
    return exp_rep


# 元素記述子を入力表現に変換する関数
def elem_rep(read_col_name: str, desc: pd.DataFrame, aw: pd.Series,
             cand_elem_num: int, cand_cols_order: dict, condition: dict,
             elem_lists: list, start: int, end: int) -> list:
    """
    Convert elemental descriptors into an input representation for machine learning.

    Parameters
    ----------
    read_col_name : str
        The name of the column to read from the catalyst combinations.
    desc : pd.DataFrame
        A DataFrame containing the element descriptors.
    aw : pd.Series
        A Series containing the atomic weights of the elements.
    cand_elem_num : int
        The number of candidate elements to be used in the representation.
    cand_cols_order : dict
        A dictionary specifying the column order for candidate elements and descriptors.
    condition : dict
        A dictionary containing various parameters for the catalyst and experimental conditions.
    elem_lists : list
        A list containing the usable elements for PGM, additives, and supports.
    start : int
        The starting index for slicing the combinations.
    end : int
        The ending index for slicing the combinations.

    Returns
    -------
    list
        A list of sparse CSR matrices, where each matrix represents the descriptors for a specific element 
        (or element position) in the catalyst combination.

    Notes
    -----
    - This function generates all possible combinations of catalyst compositions using the `catal_comb` function.
    - It slices the generated combinations based on the `start` and `end` indices, and extracts the element combinations specified by `read_col_name`.
    - If only one candidate element is provided (`cand_elem_num == 1`), it directly converts the descriptors for the single element into a sparse matrix.
    - If multiple candidate elements are provided, it processes each element position separately, normalizing the descriptors by atomic weight, and returns a list of sparse matrices.
    """
    elem_dict = {}
    read_col = cand_cols_order[read_col_name]
    all_comb, total_comb_num = catal_comb(condition, elem_lists)
    all_comb_slice = itertools.islice(all_comb, start, end)
    elem_comp_dict = [u[read_col] for u in all_comb_slice]

    if cand_elem_num == 1:
        elem_rep = [np.array(desc.loc[elem_comp_dict]) /
                    np.array(aw.loc[elem_comp_dict]).reshape(-1, 1)]
        elem_rep = [sparse.csr_matrix(m) for m in elem_rep]

    else:
        elem_dict = [np.asarray([i[n] for i in elem_comp_dict]) for n in range(cand_elem_num)]
        elem_rep = [np.array(desc.loc[elem_dict[i]]) /
                    np.array(aw.loc[elem_dict[i]]).reshape(-1, 1) for i in range(cand_elem_num)]
        elem_rep = [sparse.csr_matrix(m) for m in elem_rep]
    return elem_rep


# 上で定義した3種の関数(one_hot, exp_rep, elem_rep)を使って、EI値計算のための最終的な入力表現を作成する関数
def input_comps(condition: dict, converted: dict, elem_lists: list,
                cand_cols_order: dict, slice_size: int, start: int, end: int) -> tuple:

    pgm_model = condition['pgm_model']
    add_model = condition['add_model']
    supp_model = condition['supp_model']

    cand_pgm_num = condition['cand_pgm_num']
    cand_add_num = condition['cand_add_num']
    cand_supp_num = condition['cand_supp_num']

    CalT_num = condition['CalT_num']

    pgm_desc = converted['pgm_desc']
    add_desc = converted['add_desc']
    supp_desc = converted['supp_desc']
    
    pgm_aw = converted['pgm_aw']
    add_aw = converted['add_aw']
    supp_mw = converted['supp_mw']

    all_pgm_list = elem_lists['all_pgm']
    all_add_list = elem_lists['all_add']
    all_supp_list = elem_lists['all_supp']

    all_comb, total_comb_num = catal_comb(condition, elem_lists)
    comb_slice = itertools.islice(all_comb, start, end)
    comb_slice_list = list(comb_slice)


    if pgm_model == 0 and cand_pgm_num != 0:
        pgm_rep = None
        pgm_oh = one_hot('pgm', all_pgm_list, cand_pgm_num, cand_cols_order,
                         condition, elem_lists, slice_size, start, end)
    elif pgm_model == 1 and cand_pgm_num != 0:
        pgm_rep = elem_rep('pgm', pgm_desc, pgm_aw, cand_pgm_num,
                           cand_cols_order, condition, elem_lists, start, end)
        pgm_oh = one_hot('pgm', all_pgm_list, cand_pgm_num, cand_cols_order,
                         condition, elem_lists, slice_size, start, end)
    elif pgm_model == 2 and cand_pgm_num != 0:
        pgm_rep = elem_rep('pgm', pgm_desc, pgm_aw, cand_pgm_num,
                           cand_cols_order, condition, elem_lists, start, end)
        pgm_oh = None
    else:
        pgm_rep, pgm_oh = None, None


    if add_model == 0 and cand_add_num != 0:
        add_rep = None
        add_oh = one_hot('add', all_add_list, cand_add_num, cand_cols_order,
                         condition, elem_lists, slice_size, start, end)
    elif add_model == 1 and cand_add_num != 0:
        add_rep = elem_rep('add', add_desc, add_aw, cand_add_num,
                           cand_cols_order, condition, elem_lists, start, end)
        add_oh = one_hot('add', all_add_list, cand_add_num, cand_cols_order,
                         condition, elem_lists, slice_size, start, end)
    elif add_model == 2 and cand_add_num != 0:
        add_rep = elem_rep('add', add_desc, add_aw, cand_add_num,
                           cand_cols_order, condition, elem_lists, start, end)
        add_oh = None
    else:
        add_rep, add_oh = None, None


    if supp_model == 0 and cand_supp_num != 0:
        supp_rep = None
        supp_oh = one_hot('supp', all_supp_list, cand_supp_num,
                          cand_cols_order, condition, elem_lists, slice_size, start, end)
    elif supp_model == 1 and cand_supp_num != 0:
        supp_rep = elem_rep('supp', supp_desc, supp_mw, cand_supp_num,
                            cand_cols_order, condition, elem_lists, start, end)
        supp_oh = one_hot('supp', all_supp_list, cand_supp_num,
                          cand_cols_order, condition, elem_lists, slice_size, start, end)
    elif supp_model == 2 and cand_supp_num != 0:
        supp_rep = elem_rep('supp', supp_desc, supp_mw, cand_supp_num,
                            cand_cols_order, condition, elem_lists, start, end)
        supp_oh = None
    else:
        supp_rep, supp_oh = None, None


    if CalT_num != 0:
        CalT_rep = exp_rep('CalT', cand_cols_order,
                           condition, elem_lists, start, end)
    else:
        CalT_rep = None

    return comb_slice_list, pgm_rep, pgm_oh, \
        add_rep, add_oh, supp_rep, supp_oh, CalT_rep




# 決定木ベースの回帰モデル群（例: ランダムフォレスト）の予測値に基づいて、各サンプルの予測標準偏差を計算する関数
def pred_std_csr(X: pd.DataFrame, trees: list, predictions: float,
                 min_variance: float = 0.0) -> np.ndarray:
    """
    Calculate the standard deviation of predictions from a set of decision tree-based regression models.

    Parameters
    ----------
    X : pd.DataFrame
        The input data for which predictions and standard deviations are calculated. Each row corresponds to a sample.
    trees : list
        A list of decision tree-based models (e.g., from a random forest or gradient-boosted trees).
    predictions : float
        The predicted values (mean) for the input data from the ensemble of trees.
    min_variance : float, optional
        The minimum allowed variance for each tree, by default 0.0. Variances smaller than this value are replaced by `min_variance`.

    Returns
    -------
    np.ndarray
        An array of standard deviations for the predictions across all trees for each input sample.

    Notes
    -----
    - The function works by calculating the impurity (variance) from each tree's nodes where the input data points fall.
    - The variance from each tree is added to the squared mean prediction of that tree, and the standard deviation is computed.
    - Negative variances are clipped to zero to avoid numerical issues.
    """
    std = np.zeros(X.shape[0])
    for tree in trees:
        # 各サンプルが属するツリーノードの不純度（分散）を取得
        var_tree = tree.tree_.impurity[tree.apply(X)]
        # 最小分散の閾値を設定
        var_tree[var_tree < min_variance] = min_variance
        # 現在のツリーを用いて各サンプルの予測平均を取得
        mean_tree = tree.predict(X)
        # 分散と予測平均の二乗を加算
        std += var_tree + mean_tree ** 2
    # ツリーの数で割って平均を求める
    std /= len(trees)
    # 全体の予測値の二乗を引いて分散を計算
    std -= predictions ** 2.0
    # 分散が負にならないようにし、標準偏差を求めるために平方根を取る
    std[std < 0.0] = 0.0
    std = std ** 0.5
    return std



# Expected improvement (EI) 値を計算する関数
def EI(Reaction: str, mu: float, sigma: float, cur_min: float, cur_max: float) -> float:

    if Reaction == 'CH2OH_Batch':
        # Maximization problem
        Z = (mu - cur_max) / sigma
        ei = (mu - cur_max) * norm.cdf(Z) + sigma * norm.pdf(Z)
    else:
        # Minimization problem
        Z = (cur_min - mu) / sigma
        ei = (cur_min - mu) * norm.cdf(Z) + sigma * norm.pdf(Z)
    
    return ei



# EI値を計算した結果、上位に残った触媒組成+実験条件のリストの形を整形する関数(encodeし、barhグラフをプロットするために使用)
def make_cand_comp(cand_elem:list, condition:dict) -> np.ndarray:

    cand_pgm_num = condition['cand_pgm_num']
    cand_add_num = condition['cand_add_num']
    cand_supp_num = condition['cand_supp_num']

    CalT_num = condition['CalT_num']

    total_elem_num = cand_pgm_num + cand_add_num + cand_supp_num
    # total_elem_exp_num = total_elem_num + CalT_num + ReacT_num + WHSV_num

    cand_cols_order = cols_order(condition)

    elem_comp_all = {}
    elem_comp_dict = {}

    elem_comp_pgm = {}
    if cand_pgm_num != 0:
        for i in range(0, cand_pgm_num):
            elem_comp_dict[i] = [x[cand_cols_order['pgm']][i] for x in cand_elem]
            elem_comp_pgm[i] = np.array([x for x in elem_comp_dict[i]]).reshape(-1, 1)
        elem_comp_all['pgm'] = np.hstack([elem_comp_pgm[i] for i in range(cand_pgm_num)]).reshape(-1, cand_pgm_num)

    elem_comp_add = {}
    if cand_add_num != 0:
        for i in range(0, cand_add_num):
            elem_comp_dict[i+cand_pgm_num] = [x[cand_cols_order['add']][i] for x in cand_elem]
            elem_comp_add[i] = np.array([x for x in elem_comp_dict[i+cand_pgm_num]]).reshape(-1, 1)
        elem_comp_all['add'] = np.hstack([elem_comp_add[i] for i in range(cand_add_num)]).reshape(-1, cand_add_num)

    elem_comp_supp = {}
    if cand_supp_num != 0:
        for i in range(0, cand_supp_num):
            elem_comp_dict[i+cand_pgm_num+cand_add_num] = [x[cand_cols_order['supp']] for x in cand_elem]
            elem_comp_supp[i] = np.array([x for x in elem_comp_dict[i+cand_pgm_num+cand_add_num]]).reshape(-1, 1)
        elem_comp_all['supp'] = np.hstack([elem_comp_supp[i] for i in range(cand_supp_num)]).reshape(-1, cand_supp_num)

    elem_comp_CalT = {}
    if CalT_num != 0:
        for i in range(0, CalT_num):
            elem_comp_dict[i+total_elem_num] = [x[cand_cols_order['CalT']] for x in cand_elem]
            elem_comp_CalT[i] = np.array([x for x in elem_comp_dict[i+total_elem_num]]).reshape(-1, 1)
        elem_comp_all['CalT'] = np.hstack([elem_comp_CalT[i] for i in range(CalT_num)]).reshape(-1, CalT_num)

    cand_comp = np.hstack([elem_comp_all[i] for i in list(cand_cols_order.keys())])
    return cand_comp



# 触媒組成＋実験条件を文字列(str)として整形し、棒(barh)グラフの軸要素(yticks)を作成する関数
def encode(elem_wt: list, elem: str, condition: dict) -> str:

    cand_pgm_num = condition['cand_pgm_num']
    cand_add_num = condition['cand_add_num']
    cand_supp_num = condition['cand_supp_num']

    CalT_num = condition['CalT_num']

    total_elem_num = cand_pgm_num + cand_add_num + cand_supp_num
    total_cand_num = total_elem_num + CalT_num

    encode_name = ""
    for i in range(0, cand_pgm_num):
        if i == total_cand_num - 1:
            encode_name += f"{elem[i]} {elem_wt[i]}"
        else:
            encode_name += f"{elem[i]} {elem_wt[i]}, "

    for i in range(cand_pgm_num, cand_pgm_num+cand_add_num):
        if i == total_cand_num - 1:
            encode_name += f"{elem[i]} {elem_wt[i]}"
        else:
            encode_name += f"{elem[i]} {elem_wt[i]}, "

    for i in range(cand_pgm_num+cand_add_num, total_cand_num):
        if i == total_cand_num - 1:
            encode_name += f"{elem[i]}"
        else:
            encode_name += f"{elem[i]}, "

    return encode_name



# 上で定義した2つの関数(make_cand_comp, encode)を順番に呼び出して処理を実行する関数。
def get_encode(idx: int, elem_wt: list, all_comb_slice: list, condition: dict) -> list:
    """
    Execute the process of generating catalyst compositions and encoding them into strings for plotting.

    Parameters
    ----------
    idx : int
        Index specifying which combinations from the `all_comb_slice` to process.
    elem_wt : list
        A list of element weights corresponding to the catalyst composition.
    all_comb_slice : list
        A list containing all possible combinations of catalyst compositions and experimental conditions.
    condition : dict
        A dictionary containing the number of candidate PGMs, Promoters, Additives, and Supports elements.

    Returns
    -------
    list
        A list of encoded strings representing the catalyst composition and experimental conditions.
    
    Notes
    -----
    - This function first retrieves the catalyst compositions based on the provided indices.
    - It then calls `make_cand_comp` to format the catalyst compositions and `encode` to generate strings 
      for plotting or further analysis.
    """
    catal_comb_idx = [all_comb_slice[i] for i in idx]
    catal_comp = make_cand_comp(catal_comb_idx, condition)
    catal_name = [encode(elem_wt, elem, condition) for elem in catal_comp]
    return catal_name



# 上で定義した get_encode関数で文字列(str)に変換した触媒組成を、再び要素ごとに分解してpd.DataFrame形式で出力する関数
def cand_separation(cand_data: pd.DataFrame, condition: dict) -> pd.DataFrame:

    cand_pgm_num = condition['cand_pgm_num']
    cand_add_num = condition['cand_add_num']

    CalT_num = condition['CalT_num']

    data_cols = setting.data_columns(condition)
    cand_elem_cols = data_cols['cand_elem']
    cand_wt_cols = data_cols['cand_wt']
    cand_labels = data_cols['cand_labels']

    data = pd.read_csv(cand_data, sep=',')
    comp = data.loc[:, 'Unnamed: 0']
    ei = data.loc[:, 'ei']

    sep_data = []
    for i in range(len(data)):
        string = comp[i].split()
        string.append(ei[i])
        sep_data.append(string)

    cand_sep_data = pd.DataFrame(sep_data)
    cand_sep_data.columns = cand_labels
    print('cand_sep_data:', cand_sep_data.shape)

    sep_data_wt = []
    if cand_pgm_num != 0:
        for i in range(cand_pgm_num):
            cand_sep_data_wt_pgm = cand_sep_data[f'PGM{i+1}_wt%'].str.strip(',')
            sep_data_wt.append(cand_sep_data_wt_pgm)

    if cand_add_num != 0:
        for i in range(cand_add_num):
            cand_sep_data_wt_add = cand_sep_data[f'Ad{i+1}_wt%'].str.strip(',')
            sep_data_wt.append(cand_sep_data_wt_add)

    cand_wt = pd.DataFrame(sep_data_wt).transpose()
    cand_wt.columns = cand_wt_cols

    cand_elem = cand_sep_data.loc[:, cand_elem_cols]
    cand_ei = cand_sep_data.loc[:, 'ei']

    if CalT_num != 0:
        cand_CalT = cand_sep_data.loc[:, 'Calc T']
        cand = pd.concat([cand_elem, cand_wt, cand_CalT, cand_ei], axis=1)

    elif CalT_num == 0:
        cand = pd.concat([cand_elem, cand_wt, cand_ei], axis=1)

    cand = cand.reindex(columns=cand_labels)
    cand[cand_wt_cols] = cand[cand_wt_cols].astype(float)
    
    return cand



# 要素ごとに分解した触媒組成を、再び文字列(str)として結合し直す関数(K-cluster後plotのために使用)
def data_str(data: pd.DataFrame, condition: dict, data_cols: list) -> pd.DataFrame:

    pgm_num = condition['pgm_num']
    add_num = condition['add_num']
    supp_num = condition['supp_num']

    CalT_num = condition['CalT_num']

    elem_cols, wt_cols = data_cols['elem'], data_cols['wt']

    data['Catalyst'] = ''
    for i in range(len(data)):
        for j in range(pgm_num+add_num):
            if data[wt_cols[j]][i] > 0.00:
                data['Catalyst'][i] += data[elem_cols[j]][i] + \
                    '(' + data[wt_cols[j]][i].astype(str) + '), '
                
        if supp_num != 0:
            data['Catalyst'][i] += data['Support_name'][i]

        if CalT_num == 1:
            data['Catalyst'][i] += ', ' + \
                '(' + data['Calc T'][i].astype(str) + ')'
    return data




# 上で定義した cand_separation関数で要素ごとに分解した触媒組成を、再び文字列(str)として結合し直す関数(K-cluster後plotのために使用)
def cand_str(cand: pd.DataFrame, condition: dict, data_cols: list) -> pd.DataFrame:

    cand_pgm_num = condition['cand_pgm_num']
    cand_add_num = condition['cand_add_num']
    cand_supp_num = condition['cand_supp_num']

    CalT_num = condition['CalT_num']

    cand_elem_cols, cand_wt_cols = data_cols['cand_elem'], data_cols['cand_wt']

    cand['Top catal.'] = ''
    for i in range(len(cand)):
        for j in range(cand_pgm_num+cand_add_num):
            if cand[cand_wt_cols[j]][i] > 0.00:
                cand['Top catal.'][i] += cand[cand_elem_cols[j]][i] + \
                    '(' + cand[cand_wt_cols[j]][i].astype(str) + '), '
        if cand_supp_num != 0:
            cand['Top catal.'][i] += cand['Support_name'][i]
        if CalT_num == 1:
            cand['Top catal.'][i] += ' ' + '(' + cand['Calc T'][i] + ')'

    return cand
