#!/usr/bin/env python
import sys, os, shutil, glob
sys.dont_write_bytecode = True
from typing import List
import subprocess

import numpy as np

if os.path.isfile('conditions.py'):
    shutil.move('conditions.py', 'modules/conditions.py')

import modules.conditions as conditions
import modules.setting as setting
import modules.converter as converter

################################
your_group_ID = 'gr10414b'
################################

import argparse
parser = argparse.ArgumentParser()
parser.add_argument('-s', '--job_size', type=int, default=50)
parser.add_argument('-l', '--job_limit', type=int, default=128)
args = vars(parser.parse_args())


###################################################################
###################################################################
def divisors_list(num: int) -> List[int]:
    """
    Generate a list of divisors of a given integer in ascending order.

    Parameters
    ----------
    num : int
        The integer for which to calculate the divisors.

    Returns
    -------
    list of int
        A list containing all divisors of `num` in ascending order.
    """
    divisors = []
    for i in range(1, num + 1):
        if num % i == 0:
            divisors.append(i)
    return divisors



def num_jobs(job_num_list: List[int]) -> int:
    """
    Determine the number of jobs based on candidate job numbers and global thresholds.

    For each candidate job number in job_num_list, this function calculates a computation cost
    as the global variable `num_elem_wt` divided by the candidate job number. If the computed cost 
    exceeds the global threshold `job_size`, that candidate is selected as the number of jobs.
    If multiple candidates satisfy the condition, the last one that meets the condition is returned.

    Note:
    -----
    This function assumes that the variables `num_elem_wt` (a float) and `job_size` (a float) 
    are defined in the global scope.

    Parameters
    ----------
    job_num_list : List[int]
        A list of candidate job numbers.

    Returns
    -------
    int
        The selected number of jobs.
    """
    calc_cost = []
    num_of_jobs = None  # Will be set when condition is met.
    for i in range(len(job_num_list)):
        cost = num_elem_wt / job_num_list[i]
        calc_cost.append(cost)
        if cost > job_size:
            num_of_jobs = job_num_list[i]
    return num_of_jobs
###################################################################
###################################################################


job_size = args['job_size']
job_limit = args['job_limit']

condition = conditions.calc_condition()
computer = condition['computer']
Reaction = condition['Reaction']
Search_method = condition['Search_method']
data_cols = setting.data_columns(condition)


desc_path = 'data/Descriptors_{Reaction}.xlsx'
if os.path.exists(desc_path):
    print(f'Confirmed the existence of {desc_path}.')
else:
    os.rename(glob.glob('data/Descriptors_{Reaction}_*')[0], 'data/Descriptors_{Reaction}.xlsx')
    print(f'{desc_path} has been created.')
print('')

dir_names = ['job','log', 'out', 'err']
for dir_name in dir_names:
    if os.path.exists(dir_name) == True:
        shutil.rmtree(dir_name)
        if dir_name == 'err':
            print('dirs clean up')


core = condition['core']
threads = 1 * core

elem_wt, num_elem_wt = converter.all_wt_comb(condition)
print('num_elem_wt =', num_elem_wt)

job_num_list = divisors_list(num_elem_wt)


if len(job_num_list) == 1:
    num_of_jobs = job_num_list[0]
else:
    num_of_jobs = int(num_jobs(job_num_list))
print('num_of_jobs =', num_of_jobs)
print('job_size =', int(num_elem_wt/num_of_jobs))


for dir_name in dir_names:
    os.makedirs(dir_name, exist_ok=True)


if computer == 'B':
    for i, j in enumerate(zip(
        np.arange(0, num_elem_wt+(num_elem_wt/num_of_jobs), num_elem_wt/num_of_jobs, dtype=int),
        np.arange(num_elem_wt/num_of_jobs, num_elem_wt+(num_elem_wt/num_of_jobs), num_elem_wt/num_of_jobs, dtype=int)), 1):
        with open(f'run{computer}_{i}.sh', 'w') as f:
            print(
f'#!/bin/bash\n\
#============ Slum Options ============\n\
#SBATCH -p {your_group_ID}\n\
#SBATCH -t 168:00:00\n\
#SBATCH --rsc p=1:t={threads}:c={core}:m=500G\n\
#SBATCH -J "{Reaction}_{i}"\n\
#SBATCH --no-requeue\n\
#SBATCH -o out/out_{i}\n\
#SBATCH -e err/err_{i}\n\
#============ Shell Script ============\n\
set -x\n\
\n\
srun python {Search_method}_search.py --num {i} --from {j[0]} --to {j[1]} --cores {core} > log/log_cand_{i}',
            file=f
            )
        subprocess.call(f"sbatch run{computer}_{i}.sh", shell=True)


else:
    for i, j in enumerate(zip(
        np.arange(0, num_elem_wt+(num_elem_wt/num_of_jobs), num_elem_wt/num_of_jobs, dtype=int),
        np.arange(num_elem_wt/num_of_jobs, num_elem_wt+(num_elem_wt/num_of_jobs), num_elem_wt/num_of_jobs, dtype=int)), 1):
        subprocess.call(f"python {Search_method}_search.py --num {i} --from {j[0]} --to {j[1]} --cores {core} > log/log_cand_{i}", shell=True)

subprocess.call(f"mv run{computer}_* job/", shell=True)

