#!/usr/bin/env python
import sys, os, glob
sys.dont_write_bytecode = True

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.cluster import KMeans

import modules.conditions as conditions
import modules.setting as setting
import modules.converter as converter

import random
random.seed(1107)
np.random.seed(1107)

import warnings
warnings.filterwarnings('ignore')

plt.rcParams['font.size'] = 18
plt.rcParams['xtick.labelsize'] = 17
plt.rcParams['ytick.labelsize'] = 17
plt.rcParams['font.family'] = 'DejaVu Sans'
plt.rcParams['font.sans-serif'] = ['Arial']
plt.rcParams["mathtext.fontset"] = "stixsans"
plt.rcParams.update({'mathtext.default':'regular'})
plt.rcParams['xtick.direction'] = 'out'
plt.rcParams['ytick.direction'] = 'out'
plt.rcParams['xtick.major.width'] = 1.0
plt.rcParams['ytick.major.width'] = 1.0
plt.rcParams['axes.linewidth'] = 1.0
plt.rcParams['axes.grid'] = False
plt.rcParams['grid.linestyle'] = '--'
plt.rcParams['grid.linewidth'] = 2
plt.rcParams["legend.markerscale"] = 2
plt.rcParams["legend.fancybox"] = False
plt.rcParams["legend.framealpha"] = 1
plt.rcParams["legend.edgecolor"] = 'black'

PATH = 'results'
os.makedirs(PATH, exist_ok = True)

import argparse
parser = argparse.ArgumentParser()
parser.add_argument('-n', '--catal_number', type=int, default=100)
parser.add_argument('-k', '--K_cluster', type=int, default=100)

args = vars(parser.parse_args())

catal_number = args['catal_number']
K_cluster = args['K_cluster']
w_space = 1.8


### INPUT FILES ###
condition = conditions.calc_condition()
date, Reaction = condition['date'], condition['Reaction']
ML_model, Search_method = condition['ML_model'], condition['Search_method']

pgm_model = condition['pgm_model']
add_model = condition['add_model']
supp_model = condition['supp_model']

data_cols = setting.data_columns(condition)
desc_cols = setting.desc_columns(Reaction)

converted = converter.data_convert(condition, data_cols, desc_cols,
                                   sheet_name='data_sheet_name',
                                   select_desc=True, use_models=[],
                                   local=False, idx=None)

### Top XX catalysts OUTPUT Files ###
out_csv_file_top = f'results/{date}_{Reaction}_cand_sum_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}.csv'
fig_title_top = f'{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_top{catal_number}'
out_fig_name_top = f'results/{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_top{catal_number}.png'

out_csv_file_K = f"results/{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_K{K_cluster}.csv"
fig_title_K = f'{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_K{K_cluster}'
out_fig_name_K = f"results/{date}_{Reaction}_{ML_model}_prop{pgm_model}{add_model}{supp_model}_{Search_method}_K{K_cluster}.png"


csv_files = glob.glob('cand/*.csv')
data_list = []
for file in csv_files:
	print(file)
	data_list.append(pd.read_csv(file))

cand = pd.concat(data_list, axis=0, sort=True)
print("cand_shape: ", cand.shape)
cand.to_csv(out_csv_file_top, index=False)

cand_dict = converter.cand_convert(condition, converted, data_cols, desc_cols,
                                select_desc=True, use_models=[], local=False)
cand, feat, target = cand_dict['cand'], cand_dict['feat'], cand_dict['target']

top_catal = cand.sort_values('ei')
top_catal = top_catal.iloc[-catal_number:, :]
x_top = np.arange(len(top_catal))


if catal_number <= 50:
	plt.figure(figsize=(10,20), facecolor='white')
	plt.barh(x_top,top_catal.loc[:,'ei'], color='limegreen', alpha=0.7)
	plt.xticks(fontsize=18)
	plt.yticks(x_top, top_catal.loc[:, 'Top catal.'], fontsize=16)
	plt.title(fig_title_top, fontsize=20)
	plt.xlabel('Expected improvement (EI)', fontsize=18)
	plt.ylabel('Candidate catalyst composition', fontsize=18)
	plt.savefig(out_fig_name_top, dpi=300, bbox_inches="tight")

elif catal_number <= 100 and catal_number > 50:
	fig = plt.figure(figsize=(25,20), facecolor='white')
	ax1 = fig.add_subplot(1,2,1)
	ax2 = fig.add_subplot(1,2,2)
	fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=w_space)

	top_catal1 = top_catal.iloc[0:int(1/2*catal_number),:]
	top_catal2 = top_catal.iloc[int(1/2*catal_number):catal_number,:]
	x_top1 = np.arange(int(1/2*len(top_catal)))

	fig.suptitle(fig_title_top, x=0.35, y=1.03, fontsize=24)
	ax1.barh(x_top1, top_catal1.loc[:,'ei'], color='limegreen', alpha=0.7)
	ax1.invert_yaxis()
	ax1.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax1.tick_params(axis='x', labelsize=18)
	ax1.set_yticks(x_top1, labels=top_catal1.loc[:, 'Top catal.'], fontsize=16)
	ax1.set_ylabel('Candidate catalyst composition', fontsize=24)
	ax1.set_yticklabels(top_catal1.loc[:, 'Top catal.'], fontsize=16)

	ax2.barh(x_top1, top_catal2.loc[:,'ei'], color='limegreen', alpha=0.7)
	ax2.invert_yaxis()
	ax2.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax2.tick_params(axis='x', labelsize=18)
	ax2.set_yticks(x_top1, labels=top_catal2.loc[:, 'Top catal.'], fontsize=16)
	ax2.set_yticklabels(top_catal2.loc[:, 'Top catal.'], fontsize=16)
	plt.savefig(out_fig_name_top, dpi=300, bbox_inches="tight")

elif catal_number <= 150 and catal_number > 100:
	fig = plt.figure(figsize=(35,20), facecolor='white')
	ax1 = fig.add_subplot(1,3,1)
	ax2 = fig.add_subplot(1,3,2)
	ax3 = fig.add_subplot(1,3,3)
	fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=w_space)

	top_catal1 = top_catal.iloc[0:int(1/3*catal_number),:]
	top_catal2 = top_catal.iloc[int(1/3*catal_number):int(2/3*catal_number),:]
	top_catal3 = top_catal.iloc[int(2/3*catal_number):catal_number,:]
	x_top1 = np.arange(int(1/3*len(top_catal)))

	fig.suptitle(fig_title_top, x=0.4, y=1.03, fontsize=24)
	ax1.barh(x_top1, top_catal1.loc[:,'ei'], color='limegreen', alpha=0.7)
	ax1.invert_yaxis()
	ax1.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax1.tick_params(axis='x', labelsize=18)
	ax1.set_yticks(x_top1, labels=top_catal1.loc[:, 'Top catal.'], fontsize=16)
	ax1.set_ylabel('Candidate catalyst composition', fontsize=24)
	ax1.set_yticklabels(top_catal1.loc[:, 'Top catal.'], fontsize=16)

	ax2.barh(x_top1, top_catal2.loc[:,'ei'], color='limegreen', alpha=0.7)
	ax2.invert_yaxis()
	ax2.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax2.tick_params(axis='x', labelsize=18)
	ax2.set_yticks(x_top1, labels=top_catal2.loc[:, 'Top catal.'], fontsize=16)
	ax2.set_yticklabels(top_catal2.loc[:, 'Top catal.'], fontsize=16)

	ax3.barh(x_top1, top_catal3.loc[:,'ei'], color='limegreen', alpha=0.7)
	ax3.invert_yaxis()
	ax3.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax3.tick_params(axis='x', labelsize=18)
	ax3.set_yticks(x_top1, labels=top_catal3.loc[:, 'Top catal.'], fontsize=16)
	ax3.set_ylabel('Candidate catalyst composition', fontsize=24)
	plt.savefig(out_fig_name_top, dpi=300, bbox_inches="tight")

elif catal_number > 150:
	fig = plt.figure(figsize=(50,20), facecolor='white')
	ax1 = fig.add_subplot(1,4,1)
	ax2 = fig.add_subplot(1,4,2)
	ax3 = fig.add_subplot(1,4,3)
	ax4 = fig.add_subplot(1,4,4)
	fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=w_space)

	top_catal1 = top_catal.iloc[0:int(1/4*catal_number),:]
	top_catal2 = top_catal.iloc[int(1/4*catal_number):int(1/2*catal_number),:]
	top_catal3 = top_catal.iloc[int(1/2*catal_number):int(3/4*catal_number),:]
	top_catal4 = top_catal.iloc[int(3/4*catal_number):catal_number,:]
	x_top1 = np.arange(int(1/4*len(top_catal)))

	fig.suptitle(fig_title_top, x=0.43, y=1.03, fontsize=24)
	ax1.barh(x_top1, top_catal1.loc[:,'ei'], color='limegreen', alpha=0.7)
	ax1.invert_yaxis()
	ax1.set_xlabel('Expected improvement (EI)', fontsize=20)
	ax1.tick_params(axis='x', labelsize=18)
	ax1.set_yticks(x_top1, labels=top_catal1.loc[:, 'Top catal.'], fontsize=16)
	ax1.set_ylabel('Candidate catalyst composition', fontsize=24)
	ax1.set_yticklabels(top_catal1.loc[:, 'Top catal.'], fontsize=16)

	ax2.barh(x_top1, top_catal2.loc[:,'ei'], color='limegreen', alpha=0.7)
	ax2.invert_yaxis()
	ax2.set_xlabel('Expected improvement (EI)', fontsize=20)
	ax2.tick_params(axis='x', labelsize=18)
	ax2.set_yticks(x_top1, labels=top_catal2.loc[:, 'Top catal.'], fontsize=16)
	ax2.set_yticklabels(top_catal2.loc[:, 'Top catal.'], fontsize=16)

	ax3.barh(x_top1, top_catal3.loc[:,'ei'], color='limegreen', alpha=0.7)
	ax3.invert_yaxis()
	ax3.set_xlabel('Expected improvement (EI)', fontsize=20)
	ax3.tick_params(axis='x', labelsize=18)
	ax3.set_yticks(x_top1, labels=top_catal3.loc[:, 'Top catal.'], fontsize=16)
	ax3.set_yticklabels(top_catal3.loc[:, 'Top catal.'], fontsize=16)

	ax4.barh(x_top1, top_catal4.loc[:,'ei'], color='limegreen', alpha=0.7)
	ax4.invert_yaxis()
	ax4.set_xlabel('Expected improvement (EI)', fontsize=20)
	ax4.tick_params(axis='x', labelsize=18)
	ax4.set_yticks(x_top1, labels=top_catal4.loc[:, 'Top catal.'], fontsize=16)
	ax4.set_yticklabels(top_catal4.loc[:, 'Top catal.'], fontsize=16)
	plt.savefig(out_fig_name_top, dpi=300, bbox_inches="tight")

### PLOT K-TOP CATALYST CANDIDATES ###

# cvmodel, error = analysis.grid_search_CV(feat, target, split_type='KFold', split_num=10, use_model='ETR')
# print('n_estimators=', cvmodel.best_params_['n_estimators'], 'bootstrap=', cvmodel.best_params_['bootstrap'])
# model  = ExtraTreesRegressor(n_estimators=cvmodel.best_params_['n_estimators'], bootstrap=cvmodel.best_params_['bootstrap'], random_state=1107, n_jobs=-1)
model = ExtraTreesRegressor(n_estimators=condition['n_est'], random_state=1107, n_jobs=-1)
model.fit(feat, target)

k = KMeans(n_clusters=K_cluster, random_state=1107)
cluster = k.fit_predict(feat)
cluster = pd.Series(cluster, index=cand.index, name='cluster')

cand_top = cand.copy()
cand = pd.concat([cand, cluster], axis=1)
cand = cand.sort_values(by=['cluster','ei']).drop_duplicates(subset=['cluster'],keep='last')
cand = cand.sort_values(by='ei', ascending=False)
cand.to_csv(out_csv_file_K)
print(f'cand_K{K_cluster}_shape:', cand.shape)
# cand = converter.cand_str(cand, condition, data_cols)

K_catal = cand.sort_values('ei')
K_catal = cand.iloc[-K_cluster:, :]
x_K = np.arange(len(K_catal))

if K_cluster <= 50:
	plt.figure(figsize=(10,20), facecolor='white')
	plt.xticks(fontsize=18)
	plt.yticks(x_K, K_catal.loc[:, 'Top catal.'], fontsize=16)
	plt.barh(x_K, K_catal.loc[:,'ei'], color='blue', alpha=0.7)
	plt.gca().invert_yaxis()
	plt.title(fig_title_K, fontsize=20)
	plt.xlabel('Expected improvement (EI)', fontsize=18)
	plt.ylabel('Candidate catalyst composition', fontsize=18)
	plt.savefig(out_fig_name_K, dpi=300, bbox_inches="tight")

elif K_cluster <= 100 and K_cluster > 50:
	fig = plt.figure(figsize=(25,20), facecolor='white')
	ax1 = fig.add_subplot(1,2,1)
	ax2 = fig.add_subplot(1,2,2)
	fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=w_space)

	K_catal1 = K_catal.iloc[0:int(1/2*K_cluster),:]
	K_catal2 = K_catal.iloc[int(1/2*K_cluster):K_cluster,:]
	x_K1 = np.arange(int(1/2*len(K_catal)))

	fig.suptitle(fig_title_K, x=0.35, y=1.03, fontsize=24)
	ax1.barh(x_K1, K_catal1.loc[:,'ei'], color='blue', alpha=0.7)
	ax1.invert_yaxis()
	ax1.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax1.tick_params(axis='x', labelsize=18)
	ax1.set_yticks(x_K1, labels=K_catal1.loc[:, 'Top catal.'], fontsize=16)
	ax1.set_ylabel('Candidate catalyst composition', fontsize=24)
	ax1.set_yticklabels(K_catal1.loc[:, 'Top catal.'], fontsize=16)

	ax2.barh(x_K1, K_catal2.loc[:,'ei'], color='blue', alpha=0.7)
	ax2.invert_yaxis()
	ax2.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax2.tick_params(axis='x', labelsize=18)
	ax2.set_yticks(x_K1, labels=K_catal2.loc[:, 'Top catal.'], fontsize=16)
	#ax2.set_ylabel('Candidate catalyst composition', fontsize=24)
	ax2.set_yticklabels(K_catal2.loc[:, 'Top catal.'], fontsize=16)
	plt.savefig(out_fig_name_K, dpi=300, bbox_inches = "tight")

elif K_cluster <= 150 and K_cluster > 100:
	fig = plt.figure(figsize=(35,20), facecolor='white')
	ax1 = fig.add_subplot(1,3,1)
	ax2 = fig.add_subplot(1,3,2)
	ax3 = fig.add_subplot(1,3,3)
	fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=w_space)

	K_catal1 = K_catal.iloc[0:int(1/3*K_cluster),:]
	K_catal2 = K_catal.iloc[int(1/3*K_cluster):int(2/3*K_cluster),:]
	K_catal3 = K_catal.iloc[int(2/3*K_cluster):K_cluster,:]
	x_K1 = np.arange(int(1/3*len(K_catal)))

	fig.suptitle(fig_title_K, x=0.4, y=1.03, fontsize=24)
	ax1.barh(x_K1, K_catal1.loc[:,'ei'], color='blue', alpha=0.7)
	ax1.invert_yaxis()
	ax1.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax1.tick_params(axis='x', labelsize=18)
	ax1.set_yticks(x_K1, labels=K_catal1.loc[:, 'Top catal.'], fontsize=16)
	ax1.set_ylabel('Candidate catalyst composition', fontsize=24)
	ax1.set_yticklabels(K_catal1.loc[:, 'Top catal.'], fontsize=16)

	ax2.barh(x_K1, K_catal2.loc[:,'ei'], color='blue', alpha=0.7)
	ax2.invert_yaxis()
	ax2.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax2.tick_params(axis='x', labelsize=18)
	ax2.set_yticks(x_K1, labels=K_catal2.loc[:, 'Top catal.'], fontsize=16)
	ax2.set_yticklabels(K_catal2.loc[:, 'Top catal.'], fontsize=16)

	ax3.barh(x_K1, K_catal3.loc[:,'ei'], color='blue', alpha=0.7)
	ax3.invert_yaxis()
	ax3.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax3.tick_params(axis='x', labelsize=18)
	ax3.set_yticks(x_K1, labels=K_catal3.loc[:, 'Top catal.'], fontsize=16)
	#ax3.set_ylabel('Candidate catalyst composition', fontsize=14)
	ax3.set_yticklabels(K_catal3.loc[:, 'Top catal.'], fontsize=16)
	plt.savefig(out_fig_name_K, dpi=300, bbox_inches = "tight")

elif K_cluster > 150:
	fig = plt.figure(figsize=(50,20), facecolor='white')
	ax1 = fig.add_subplot(1,4,1)
	ax2 = fig.add_subplot(1,4,2)
	ax3 = fig.add_subplot(1,4,3)
	ax4 = fig.add_subplot(1,4,4)
	fig.subplots_adjust(bottom=0, left=0, top=1, right=1, wspace=w_space)

	K_catal1 = K_catal.iloc[0:int(1/4*K_cluster),:]
	K_catal2 = K_catal.iloc[int(1/4*K_cluster):int(1/2*K_cluster),:]
	K_catal3 = K_catal.iloc[int(1/2*K_cluster):int(3/4*K_cluster),:]
	K_catal4 = K_catal.iloc[int(3/4*K_cluster):K_cluster,:]
	x_K1 = np.arange(int(1/4*len(K_catal)))

	fig.suptitle(fig_title_K, x=0.43, y=1.03, fontsize=14)
	ax1.barh(x_K1, K_catal1.loc[:,'ei'], color='blue', alpha=0.7)
	ax1.invert_yaxis()
	ax1.set_xlabel('Expected improvement (EI)', fontsize=20)
	ax1.tick_params(axis='x', labelsize=18)
	ax1.set_yticks(x_K1, labels=K_catal1.loc[:, 'Top catal.'], fontsize=16)
	ax1.set_ylabel('Candidate catalyst composition', fontsize=24)
	ax1.set_yticklabels(K_catal1.loc[:, 'Top catal.'], fontsize=16)

	ax2.barh(x_K1, K_catal2.loc[:,'ei'], color='blue', alpha=0.7)
	ax2.invert_yaxis()
	ax2.set_xlabel('Expected improvement (EI)', fontsize=20)
	ax2.tick_params(axis='x', labelsize=18)
	ax2.set_yticks(x_K1, labels=K_catal2.loc[:, 'Top catal.'], fontsize=16)
	ax2.set_yticklabels(K_catal2.loc[:, 'Top catal.'], fontsize=16)

	ax3.barh(x_K1, K_catal3.loc[:,'ei'], color='blue', alpha=0.7)
	ax3.invert_yaxis()
	ax3.set_xlabel('Expected improvement (EI)', fontsize=24)
	ax3.tick_params(axis='x', labelsize=18)
	ax3.set_yticks(x_K1, labels=K_catal3.loc[:, 'Top catal.'], fontsize=16)
	ax3.set_yticklabels(K_catal3.loc[:, 'Top catal.'], fontsize=16)

	ax4.barh(x_K1, K_catal4.loc[:,'ei'], color='blue', alpha=0.7)
	ax4.invert_yaxis()
	ax4.set_xlabel('Expected improvement (EI)', fontsize=20)
	ax4.tick_params(axis='x', labelsize=18)
	ax4.set_yticks(x_K1, labels=K_catal4.loc[:, 'Top catal.'], fontsize=16)
	ax4.set_yticklabels(K_catal4.loc[:, 'Top catal.'], fontsize=16)
	plt.savefig(out_fig_name_K, dpi=300, bbox_inches = "tight")
